import logging
import torch
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
from typing import Any, Dict, Optional, Tuple

from .solver import BaseSolver
from ..utils import fixpoint_iteration

logger = logging.getLogger(__name__)

class ADMMSolver(BaseSolver):
    r"""
    ADMM Consensus Solver on a single GPU.

    Args:
        num_epoch_ab (int, optional): Number of epochs for solving A and B. Defaults to 1000.
        num_epoch_cd (int, optional): Number of epochs for solving C and D. Defaults to 100.
        lambda_y (float, optional): Lasso regularization parameter for Y. Defaults to 1.0.
        lambda_z (float, optional): Lasso regularization parameter for Z. Defaults to 1.0.
        rho_ab (float, optional): ADMM's rho parameter for A and B. Defaults to 10.0.
        rho_cd (float, optional): ADMM's rho parameter for C and D. Defaults to 10.0.
        batch_feature_size (int, optional): Number of columns to solve in each solving iteration. This is used to control the memory usage.
                The solver is performed (total_rows // batch_feature_size + 1) times. Defaults to 100.
        regen_states (bool, optional): Whether to regenerate states. Defaults to False.
        tol (float, optional): Tolerance for zeroing out weights. Defaults to 1e-6.
    """
    def __init__(
        self, 
        num_epoch_ab : int = 1000,
        num_epoch_cd : int = 100,
        lambda_y : float = 1e-6,
        lambda_z : float = 1e-6,
        rho_ab : float = 10.0,
        rho_cd : float = 10.0,
        batch_feature_size : int = 100,
        regen_states : bool = False,
        tol : float = 1e-6,
    ):
        self.num_epoch_ab = num_epoch_ab
        self.num_epoch_cd = num_epoch_cd
        self.lambda_y = lambda_y
        self.lambda_z = lambda_z
        self.rho_ab = rho_ab
        self.rho_cd = rho_cd
        self.batch_feature_size = batch_feature_size
        self.regen_states = regen_states
        self.tol = tol

    def solve(
        self, 
        X : np.ndarray, 
        U : np.ndarray, 
        Z : np.ndarray, 
        Y : np.ndarray,
        model_config : Dict[str, Any],
        plot_loss : Optional[bool] = False,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve State-driven Implicit Model using the ADMM consensus algorithm.

        Args:
            X (np.ndarray): Post-activation array.
            U (np.ndarray): Input array.
            Z (np.ndarray): Pre-activation array.
            Y (np.ndarray): Output array.
            model_config (Dict[str, Any]): Model configuration.
                - activation_fn (Callable): Activation function used by the implicit model.
                - device (str): Implicit model's device.
                - atol (float): Equilibrium function's tolerance.
                - kappa (float): Wellposedness condition parameter.
            plot_loss (Optional[bool], optional): Whether to plot loss curve. Useful for tuning hyperparameters and debugging. Defaults to False.

        Returns:
            A, B, C, D (np.ndarray): Implicit model's parameters.
        """
        n, m, p, q = X.shape[0], X.shape[1], U.shape[0], Y.shape[0]

        X, U, Z, Y = torch.tensor(X), torch.tensor(U), torch.tensor(Z), torch.tensor(Y)

        logger.info("===== Start solving A and B =====")
        AB = self.parallel_solve_matrix(torch.hstack([X.T, U.T]), Z.T, is_y=False, n=n, kappa=model_config['kappa'], plot_loss=plot_loss, device=model_config['device'])
        A = AB[:, :n]
        B = AB[:, n:]

        if self.regen_states:
            X = fixpoint_iteration(A, B, U, model_config['activation_fn'], model_config['device'], atol=model_config['atol']).cpu()

        logger.info("===== Start solving C and D =====")
        CD = self.parallel_solve_matrix(torch.hstack([X.T, U.T]), Y.T, is_y=True, n=n, kappa=model_config['kappa'], plot_loss=plot_loss, device=model_config['device'])
        C = CD[:, :n]
        D = CD[:, n:]

        return A, B, C, D


    def parallel_solve_matrix(self, X, Y, is_y, n, kappa, plot_loss, device):
        """
        Dividing the data matrix for memory controlled solve.
        """
        total_rows = Y.shape[1]
        batch_rows_length = self.batch_feature_size
        num_batches = total_rows // (batch_rows_length) + 1

        W = None
        loss = 0.0
        for k in range(num_batches):

            logger.info(f"Solving batch feature {k+1}/{num_batches}")

            start_idx = k * batch_rows_length
            end_idx = min((k + 1) * batch_rows_length, total_rows)
            Y_batch = Y[:, start_idx:end_idx]
            
            W_k, loss_k = self.run_solve_opt_problem(X, Y_batch, is_y, n, k, kappa, plot_loss=plot_loss, device=device)
            
            W = np.vstack([W, W_k]) if W is not None else W_k

            loss += loss_k
        
        logger.info(f"Total Lasso loss: {loss}")
        
        return W


    def run_solve_opt_problem(self, X, Y, is_y, n, k, kappa, plot_loss, device):
        """
        ADMM Solve Wrapper.
        """
        if is_y:
            num_epoch = self.num_epoch_cd
            rho = self.rho_cd
            lambda_yz = self.lambda_y
        else:
            num_epoch = self.num_epoch_ab
            rho = self.rho_ab
            lambda_yz = self.lambda_z

        if is_y:
            admm = ADMM_CD(X.shape[1], Y.shape[1], rho, lambda_yz, device=device)
        else:
            admm = ADMM_AB(X.shape[1], Y.shape[1], n, rho, lambda_yz, kappa, device=device)

        losses = []
        with torch.no_grad():
            for i in tqdm(range(num_epoch), desc="Training Epochs"):
                admm.step(X, Y)

                loss = admm.LassoObjective(X, Y)
                tqdm.write(f"Loss: {loss}")

                losses.append(loss)

        # Plot losses 
        if plot_loss:
            plt.figure()
            plt.plot(losses)
            plt.xlabel("Epoch")
            plt.ylabel("Lasso Objective")
            plt.yscale("log")
            plt.title(f"Training Loss")
            plt.savefig(f"loss_k_{k}_isy_{is_y}.png")

            # Save the loss trace
            np.save(f"loss_trace_k_{k}_isy_{is_y}.npy", losses)

        if is_y:
            result = admm.CD.T.clone().detach().cpu().numpy()
        else:
            result = admm.avg.T.clone().detach().cpu().numpy()

        result[np.abs(result) <= self.tol] = 0

        return result, losses[-1]


class ADMM_AB:
    def __init__(
        self, 
        D : int, 
        Q : int, 
        n : int, 
        rho : float, 
        lambda_yz : float, 
        kappa : float, 
        device : str
    ):
        """
        Solver class to solve A and B (concatenated) using consensus ADMM.

        Args:
            D (int): Number of features.
            Q (int): Number of samples.
            n (int): Number of rows of A.
            rho (float): ADMM's rho parameter.
            lambda_yz (float): Lasso regularization parameter.
            kappa (float): Wellposedness condition parameter.
            device (str): Device.
        """
        self.D = D
        self.Q = Q
        self.n = n
        self.device = device
        
        self.nu_AB = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # ADMM's dual variable for AB
        self.nu_Z = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # ADMM's dual variable for Z
        self.nu_M = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # ADMM's dual variable for M

        self.rho = rho

        self.AB = torch.randn(self.D, self.Q, device=device, requires_grad=False) # running estimate of AB
        self.Z = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # Z
        self.M = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # M
        self.avg = torch.zeros(self.D, self.Q, device=device, requires_grad=False) # Average of AB, Z, M => Final solution of AB

        self.lambda_yz = lambda_yz
        self.kappa = kappa

    @torch.no_grad()
    def step(self, X, y):
        """
        ADMM's update step.
        """
        X = X.to(self.device)
        y = y.to(self.device)

        t1 = X.T.matmul(X) + self.rho * torch.eye(self.D, device=self.device)
        t2 = X.T.matmul(y) + self.rho * (self.avg - self.nu_AB)
        self.AB = torch.linalg.solve(t1, t2)

        self.Z = torch.sign(self.avg - self.nu_Z) * torch.clamp(torch.abs(self.avg - self.nu_Z) - self.lambda_yz / self.rho, min=0)

        self.M[:self.n,:] = self.project_w((self.avg - self.nu_M)[:self.n,:].T).T
        self.M[self.n:,:] = (self.avg - self.nu_M)[self.n:,:]

        self.avg = (self.AB + self.Z + self.M) / 3

        self.nu_AB = self.nu_AB + (self.AB - self.avg)
        self.nu_Z = self.nu_Z + (self.Z - self.avg)
        self.nu_M = self.nu_M + (self.M - self.avg)

    def project_w(self, matrix):
        """
        Project the matrix to the L1 norm ball.
        """
        A_np = matrix.clone().cpu().numpy()
        v = self.kappa
        x = np.abs(A_np).sum(axis=-1)

        for idx in np.where(x > v)[0]:
            a_orig = A_np[idx, :]
            a_sign = np.sign(a_orig)
            a_abs = np.abs(a_orig)
            a = np.sort(a_abs)

            s = np.sum(a) - v
            l = float(len(a))
            for i in range(len(a)):
                if s / l > a[i]:
                    s -= a[i]
                    l -= 1
                else:
                    break
            alpha = s / l if l > 0 else np.max(a_abs)
            a = a_sign * np.maximum(a_abs - alpha, 0)
            # assert np.isclose(np.abs(a).sum(), v)
            A_np[idx, :] = a

        proj = torch.tensor(A_np, dtype=self.avg.dtype, device=self.device)

        return proj

    @torch.no_grad()
    def LassoObjective(self, X, y):
        """
        Evaluate the Lasso objective.
        """
        X = X.to(self.device)
        y = y.to(self.device)
        return (0.5 * torch.norm(X.matmul(self.avg) - y)**2 + self.lambda_yz * torch.sum(torch.abs(self.avg))).item()


class ADMM_CD:
    def __init__(
        self, 
        D : int, 
        Q : int, 
        rho : float, 
        lambda_yz : float, 
        device : str
    ):
        """
        Solver class to solve C and D (concatenated) using ADMM.

        Args:
            D (int): Number of features.
            Q (int): Number of samples.
            rho (float): ADMM's rho parameter.
            lambda_yz (float): Lasso regularization parameter.
            device (str): Device.
        """
        self.D = D
        self.Q = Q
        self.device = device
        
        self.nu = torch.zeros(self.D, self.Q, device=device)
        self.rho = rho

        self.CD = torch.randn(self.D, self.Q, device=device)
        self.Z = torch.zeros(self.D, self.Q, device=device)
        self.lambda_yz = lambda_yz

    @torch.no_grad()
    def step(self, X, y):
        """
        ADMM's update step.
        """
        X = X.to(self.device)
        y = y.to(self.device)

        t1 = X.T.matmul(X) + self.rho * torch.eye(self.D, device=self.device)
        t2 = X.T.matmul(y) + self.rho * self.Z - self.nu
        self.CD = torch.linalg.solve(t1, t2)

        self.Z = self.CD + self.nu / self.rho - (self.lambda_yz / self.rho) * torch.sign(self.Z).to(self.device)
        self.nu = self.nu + self.rho * (self.CD - self.Z)

    @torch.no_grad()
    def LassoObjective(self, X, y):
        """
        Evaluate the Lasso objective.
        """
        X = X.to(self.device)
        y = y.to(self.device)
        return (0.5 * torch.norm(X.matmul(self.CD) - y)**2 + self.lambda_yz * torch.sum(torch.abs(self.CD))).item()



from .spiky_data import spiky_synthetic_dataset

def load_data(args):
    """
    Load training and testing data loaders for the specified dataset.
    """

    # Spiky Synthetic Dataset
    if args.dataset == 'spiky':
        x_train, x_test, y_train, y_test = spiky_synthetic_dataset(args.look_back)
    
    else:
        raise ValueError(f"Unsupported dataset: {args.dataset}")
    
    return x_train, x_test, y_train, y_test


from .implicit_base_model import ImplicitModel
import torch
from torch import nn, Tensor
from typing import Any, Tuple


class ImplicitRNNCell(nn.Module):
    def __init__(
        self,
        input_dim: int,
        implicit_hidden_dim: int,
        hidden_dim: int,
        **kwargs: Any
    ) -> None:
        """
        An RNN cell that utilizes an implicit model for recurrent operations.
        
        Mimics a vanilla RNN, but where the recurrent operation is performed by an implicit model instead of a linear layer.
        Follows the "batch first" convention, where the input x has shape (batch_size, seq_len, input_dim).
        The hidden state doubles as the output, but it is not recommended to use it directly as the model's prediction
        for low-dimensionality problems, since it is responsible for passing information between timesteps. Instead, it is
        recommended to use a larger hidden state with a linear layer that scales it down to the desired output dimension.
        
        Args:
            input_dim (int): Dimensionality of the input features.
            implicit_hidden_dim (int): Number of hidden features in the underlying implicit model.
            hidden_dim (int): Dimensionality of the recurrent hidden state.
            **kwargs: Additional keyword arguments for the underlying ImplicitModel.
        """
        
        super().__init__()
        self.hidden_dim: int = hidden_dim
        self.layer: ImplicitModel = ImplicitModel(
            hidden_dim=implicit_hidden_dim, 
            input_dim=input_dim + hidden_dim, 
            output_dim=hidden_dim, 
            **kwargs
        )

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Forward pass of the ImplicitRNNCell.

        Args:
            x (Tensor): Input tensor of shape (batch_size, seq_len, input_dim).

        Returns:
            Tuple[Tensor, Tensor]:
                - outputs (Tensor): Hidden states for all timesteps, shape (batch_size, seq_len, hidden_dim).
                - h (Tensor): Final hidden state, shape (batch_size, hidden_dim).
        """

        outputs = torch.empty(*x.shape[:-1], self.hidden_dim, device=x.device)
        
        h = torch.zeros(x.shape[0], self.hidden_dim, device=x.device)
        for t in range(x.shape[1]):
            h = self.layer(torch.cat((x[:, t, :], h), dim=-1))
            outputs[:, t, :] = h
            
        return outputs, h
    

class ImplicitRNN(nn.Module):
    r""" 
    Implicit Recurrent Neural Networks.
    
    Args:
        input_dim (int): Number of input features (:math:`p`).
        output_dim (int): Number of output features (:math:`q`).
        implicit_hidden_dim (int): Hidden dimension in the implicit layer (:math:`n`).
        hidden_dim (int): Size of the recurrent hidden state (:math:`n`).
        **kwargs (Any): Additional keyword arguments for `ImplicitModel`.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        implicit_hidden_dim: int,
        hidden_dim: int,
        **kwargs: Any
    ) -> None:
        
        super().__init__()
        self.recurrent: ImplicitRNNCell = ImplicitRNNCell(
            input_dim=input_dim,
            implicit_hidden_dim=implicit_hidden_dim,
            hidden_dim=hidden_dim,
            **kwargs
        )
        self.linear: nn.Linear = nn.Linear(hidden_dim, output_dim)
        
    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of `ImplicitRNN`.

        Args:
            x (Tensor): Input tensor of shape (batch_size, seq_len, input_dim).

        Returns:
            Tensor: Output tensor of shape (batch_size, output_dim).
        """

        _, h = self.recurrent(x)
        return self.linear(h)



"""

This file contains the code for a simple feed forward network used in the experiments.

"""

import torch
from torch import nn
import numpy as np


class FashionMNIST_FFNN(nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(input_size, 64, bias=False),
            nn.ReLU(),
            nn.Linear(64, 32, bias=False),
            nn.ReLU(),
            nn.Linear(32, 16, bias=False),
            nn.ReLU(),
            nn.Linear(16, output_size, bias=False),
        )

    def forward(self, x):
        if x.ndim == 4:
            x = x.squeeze(1).flatten(start_dim=-2)
        return self.linear_relu_stack(x)
        # return nn.functional.softmax(self.linear_relu_stack(x), dim=-1)

    def scale_network(self, factor=0.99):
        layers_indices = [0, 2, 4]

        max_norm = max(
            torch.linalg.norm(self.linear_relu_stack[i].weight, np.inf)
            for i in layers_indices
        )

        for i in layers_indices:
            weight = self.linear_relu_stack[i].weight
            scaled_weight = torch.nn.Parameter(weight / (max_norm * factor))
            self.linear_relu_stack[i].weight = scaled_weight

        scaled_norm = max(
            torch.linalg.norm(self.linear_relu_stack[i].weight, np.inf)
            for i in layers_indices
        )
        print(f"Original norm : {max_norm}, Scaled norm: {scaled_norm}")

        return self



"""
Utility functions for SIM
"""

import logging

import gc
import numpy as np
import torch
import torch.nn.functional as F

logger = logging.getLogger(__name__)


def csr_to_sparse_tensor(csr, device):
    """
    Convert scipy csr_matrix to torch.sparse_coo_tensor
    """
    coordinate_tensor = csr.tocoo()
    indices = np.vstack((coordinate_tensor.row, coordinate_tensor.col))
    values = coordinate_tensor.data
    i = torch.LongTensor(indices)
    v = torch.FloatTensor(values)
    return torch.sparse_coo_tensor(i, v, csr.shape).to(device)


@torch.no_grad()
def fixpoint_iteration(A, B, U, activation_fn, device, max_iter=50000, atol=1e-6, patience=1000):
    """
    Compute X = phi(A @ X + B @ U) until convergence and return X, with support for sparse matrices.
    """

    # Convert scipy sparse matrices to PyTorch sparse tensors
    if isinstance(A, np.ndarray):
        A=torch.tensor(A,device=device)
        B=torch.tensor(B,device=device)
    else:
        A = csr_to_sparse_tensor(A, device)
        B = csr_to_sparse_tensor(B, device)

    # Compute the infinite norm of matrix A
    inf_norm_A = torch.linalg.norm(A.to_dense(), np.inf)
    logger.info(f"Infinite norm of A: {inf_norm_A}")

    # Convert numpy array U to PyTorch tensor
    if not torch.is_tensor(U):
        U = torch.from_numpy(U).float().to(device, dtype=B.dtype)
    else:
        U = U.float().to(device, dtype=B.dtype)

    # Initialize X with zeros on GPU
    n, m = A.shape[-1], U.shape[-1]
    X = torch.zeros((n, m), device=device, dtype=A.dtype)
    next_X = activation_fn(torch.sparse.mm(A, X) + torch.sparse.mm(B, U))

    min_diff = np.inf
    iteration = 0
    while True:
        X = next_X.clone()
        next_X = activation_fn(torch.sparse.mm(A, next_X) + torch.sparse.mm(B, U))

        # Compute the infinite norm of the difference
        difference = torch.linalg.norm(next_X - X, np.inf).item()
        if difference < atol:
            logger.info(f"Fixpoint iteration converged at Iteration {iteration}")
            break

        if iteration % 1000 == 0:
            logger.info(f"Iteration {iteration}: Difference = {difference}")

        # Early stopping condition
        if difference >= min_diff:
            if patience <= 0:
                logger.info(f"Early stopping at Iteration {iteration}: Difference = {difference}")
                break
            patience -= 1
        else:
            min_diff = difference
            patience = 1000  # reset patience if there is improvement

        if iteration >= max_iter:
            logger.info("Fixpoint iteration did not converge")
            break

        iteration += 1

    return next_X



import logging
import numpy as np
from typing import Any, Dict, Tuple

from .solver import BaseSolver
from ..utils import fixpoint_iteration

logger = logging.getLogger(__name__)

class LeastSquareSolver(BaseSolver):
    r"""
    Solve using numpy.linalg.lstsq. 
    Note: This solver is fast but it cannot handle the wellposeness condition (norm(A) <= kappa).

    Args:
        regen_states (bool, optional): Whether to regenerate state data to solve exact C,D after solving A,B. Defaults to False.
        tol (float, optional): Zero out weights that are less than tol. Defaults to 1e-6.
    """
    def __init__(
        self, 
        regen_states : bool = False,
        tol : float = 1e-6,
    ):
        self.regen_states = regen_states
        self.tol = tol

    def solve(
        self, 
        X : np.ndarray, 
        U : np.ndarray, 
        Z : np.ndarray, 
        Y : np.ndarray, 
        model_config : Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve with numpy.linalg.lstsq to get an implicit model.

        Args:
            X (np.ndarray): Post-activation array.
            U (np.ndarray): Input array.
            Z (np.ndarray): Pre-activation array.
            Y (np.ndarray): Output array.
            model_config (Dict[str, Any]): Model configuration. 
                - activation_fn (Callable): Activation function used by the implicit model.
                - device (str): Implicit model's device.
                - atol (float): Equilibrium function's tolerance.
                - kappa (float): Wellposedness condition parameter.
                
        Returns:
            A, B, C, D (np.ndarray): Implicit model's parameters.
        """
        n, m, p, q = X.shape[0], X.shape[1], U.shape[0], Y.shape[0]

        logger.info("===== Start solving A and B =====")
        AB = self.solve_matrix(np.hstack([X.T, U.T]), Z.T)
        A = AB[:, :n]
        B = AB[:, n:]

        if self.regen_states:
            X = fixpoint_iteration(A, B, U, model_config['activation_fn'], model_config['device'], atol=model_config['atol']).cpu()

        logger.info("===== Start solving C and D =====")
        CD = self.solve_matrix(np.hstack([X.T, U.T]), Y.T)
        C = CD[:, :n]
        D = CD[:, n:]

        return A, B, C, D

    def solve_matrix(self, X : np.ndarray, Y : np.ndarray) -> np.ndarray:
        """
        Wrapper to solve a least square problem.

        Args:
            X (np.ndarray): Input array.
            Y (np.ndarray): Output array.
        Returns:
            W (np.ndarray): Weight array.
        """
        
        W, c, r, _ = np.linalg.lstsq(X, Y, rcond=None)

        loss = np.mean(np.square(X @ W - Y))
        
        logger.info(f"Total Lasso loss: {loss}")
        logger.info(f"Data rank: {r}")

        W[np.abs(W) <= self.tol] = 0
        
        return W.T


import argparse
import torch
import torch.optim as optim
import torch.nn.functional as F
from .load_data import load_data
from .utils import train, set_seed
import os
from idl import ImplicitModel

def parse_args():
    parser = argparse.ArgumentParser(description="Train an Implicit Base model on MNIST or CIFAR-10")
    
    # Dataset
    parser.add_argument('--dataset', type=str, choices=['mnist', 'cifar10'], required=True, help="Dataset to use: 'mnist' or 'cifar10'")
    
    # Optimization parameters
    parser.add_argument('--epochs', type=int, default=10, help="Number of epochs for training")
    parser.add_argument('--batch_size', type=int, default=100, help="Batch size for training and testing")
    parser.add_argument('--device', type=int, default=0, help="Specify the device id (e.g., 0 for cuda:0)")
    parser.add_argument('--lr', type=float, default=5e-3, help="Learning rate for the optimizer")
    parser.add_argument("--seed", type=int, default=0, help="Random seed for reproducibility")
    
    # Implicit model parameters
    parser.add_argument('--hidden_dim', type=int, default=None, help="Hidden dimension of Implicit model")
    parser.add_argument('--mitr', type=int, default=300, help="Max iterations")
    parser.add_argument('--grad_mitr', type=int, default=300, help="Max gradient iterations")
    parser.add_argument('--tol', type=float, default=3e-6, help="Tolerance for convergence")
    parser.add_argument('--grad_tol', type=float, default=3e-6, help="Gradient tolerance for convergence")
    parser.add_argument('--is_low_rank', type=bool, default=False, help="Enable low-rank configuration for the model")
    parser.add_argument('--rank', type=int, default=1, help="The number of rank for low-rank configuration")
    parser.add_argument('--kappa', type=float, default=0.99, help="Inf ball")
    
    return parser.parse_args()

def main():
    """
    Main function to train the Implicit Base model.
    """

    args = parse_args()
    set_seed(args.seed)
    train_loader, test_loader = load_data(args)

    if args.hidden_dim is None:
        raise ValueError("Error: 'hidden_dim' must be specified for the model.")

    if args.dataset == 'mnist':
        input_dim = 784 
        output_dim = 10
    
    elif args.dataset == 'cifar10':
        input_dim = 3072 
        output_dim = 10 

    # Initialize the Implicit Base model
    model = ImplicitModel(
        input_dim=input_dim, 
        output_dim=output_dim,
        hidden_dim=args.hidden_dim, 
        is_low_rank=args.is_low_rank,
        rank=args.rank,
        kappa=args.kappa,
        mitr=args.mitr, 
        grad_mitr=args.grad_mitr, 
        tol=args.tol, 
        grad_tol=args.grad_tol
        )

    # Define loss function, optimizer, device and log directory
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    loss_fn = F.cross_entropy
    device = torch.device(f"cuda:{args.device}" if torch.cuda.is_available() else "cpu")
    log_dir = os.path.join("results", f"im_{args.dataset}_{args.hidden_dim}")
    
    # Train the model
    model, log_file = train(
        args=args,
        model=model,
        train_loader=train_loader,
        test_loader=test_loader,
        optimizer=optimizer,
        loss_fn=loss_fn,
        num_epochs=args.epochs,
        log_dir=log_dir,
        device=device
    )
    print(f"Training complete. Logs saved to {log_file}")

if __name__ == '__main__':
    main()



import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
from typing import Any, Dict, Tuple, Optional

from .solver import BaseSolver
from ..utils import fixpoint_iteration

logger = logging.getLogger(__name__)

class LowRankModel(torch.nn.Module):
    def __init__(self, n, p, rank, kappa, device):
        """
        Low-rank model for solving A,B.
        """
        super(LowRankModel, self).__init__()
        self.device = device
        self.kappa = kappa

        if rank is None:
            rank = n
        elif rank > n:
            logger.warning("Rank is set to be larger than max rank, setting it back to full rank")
            rank = n

        self.L = nn.Parameter(torch.randn(n, rank, device=device))
        self.R = nn.Parameter(torch.randn(n, rank, device=device))
        self.B = nn.Parameter(torch.randn(n, p, device=device))

    def forward(self, X, U):
        """
        Forward pass with low-rank decomposition components.
        """
        X = X.to(self.device)
        U = U.to(self.device)
        output = self.L @ (self.R.T @ X) + self.B @ U
        return output

    def project_LR(self):
        """
        Project the low-rank components to satisfy the well-posedness condition.
        """
        self.L.data = self.project_w(self.L, self.kappa)
        self.R.data = self.project_w(self.R.T, self.kappa).T

    def project_w(self, matrix, v=0.99):
        """
        Project the matrix to the l1 norm ball.
        """
        A_np = matrix.detach().clone().cpu().numpy()
        x = np.abs(A_np).sum(axis=-1)

        for idx in np.where(x > v)[0]:
            a_orig = A_np[idx, :]
            a_sign = np.sign(a_orig)
            a_abs = np.abs(a_orig)
            a = np.sort(a_abs)

            s = np.sum(a) - v
            l = float(len(a))
            for i in range(len(a)):
                if s / l > a[i]:
                    s -= a[i]
                    l -= 1
                else:
                    break
            alpha = s / l if l > 0 else np.max(a_abs)
            a = a_sign * np.maximum(a_abs - alpha, 0)
            # assert np.isclose(np.abs(a).sum(), v)
            A_np[idx, :] = a

        proj = torch.tensor(A_np, dtype=matrix.dtype, device=matrix.device)

        return proj

class Trainer:
    def __init__(self, num_epoch, lambda_z, lr, verbose_epoch):
        """
        Initialize the trainer.
        """
        self.num_epoch = num_epoch
        self.lamb = lambda_z
        self.lr = lr
        self.verbose_epoch = verbose_epoch

    def train(self, model, X, U, Z):
        """
        Train the model.
        """
        optimizer = torch.optim.Adam(model.parameters(), lr=self.lr)
        losses = []
        for epoch in tqdm(range(self.num_epoch)):
            optimizer.zero_grad()
            output = model(X, U)
            loss = F.mse_loss(output, Z.to(output.device)) + self.lamb * torch.norm(model.L, p=2) + self.lamb * torch.norm(model.R, p=2)
            loss.backward()
            optimizer.step()
            model.project_LR()
            losses.append(loss.item())
            if epoch % self.verbose_epoch == 0:
                logger.info(f"Loss at epoch {epoch}: {loss.item()}")
        return model, losses

class ProjectedGDLowRankSolver(BaseSolver):
    r"""
    Train State-driven Implicit Model using projected gradient descent to force A low-rank and well-posed.
    A, B are solved using projected gradient descent.
    C, D are solved using numpy least square solver.

    Args:
        rank (int): Rank of the A.
        num_epoch (int): Number of epochs to train A, B.
        lambda_z (float): Lasso regularization parameter for Z.
        lr (float): Learning rate.
        verbose_epoch (int): Number of epochs to print the loss.
        regen_states (bool, optional): Whether to regenerate states. Defaults to False.
        tol (float, optional): Tolerance for zeroing out weights. Defaults to 1e-6.
    """
    def __init__(
        self,
        rank : Optional[int] = None,
        num_epoch : int = 10000,
        lambda_z : float = 1e-6,
        lr : float = 1e-3,
        verbose_epoch : int = 100,
        regen_states : bool = False,
        tol : float = 1e-6,
    ):
        self.rank = rank
        self.num_epoch = num_epoch
        self.lambda_z = lambda_z
        self.lr = lr
        self.verbose_epoch = verbose_epoch
        self.regen_states = regen_states
        self.tol = tol

    def solve(
        self,
        X : np.ndarray,
        U : np.ndarray,
        Z : np.ndarray,
        Y : np.ndarray,
        model_config : Dict[str, Any],
        plot_loss : bool = False,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve the implicit model and force A to be low-rank.
        
        Args:
            X (np.ndarray): Post-activation array.
            U (np.ndarray): Input array.
            Z (np.ndarray): Pre-activation array.
            Y (np.ndarray): Output array.
            model_config (Dict[str, Any]): Implicit model configuration.
                - activation_fn (Callable): Activation function used by the implicit model.
                - device (str): Implicit model's device.
                - atol (float): Equilibrium function's tolerance.
                - kappa (float): Wellposedness condition parameter.
            plot_loss (bool, optional): Whether to plot the loss. Defaults to False.
        
        Returns:
            A, B, C, D (np.ndarray): Implicit model's parameters.
        """
        n, m, p, q = X.shape[0], X.shape[1], U.shape[0], Y.shape[0]

        X, U, Z = torch.tensor(X), torch.tensor(U), torch.tensor(Z)

        logger.info("===== Start parallel solve for A and B =====")
        L, R, B = self.lowrank_solve_matrix(X, U, Z, n=n, p=p, kappa=model_config["kappa"], device=model_config["device"], plot_loss=plot_loss)
        A = L @ R.T
        logger.info(f"Rank A: {np.linalg.matrix_rank(A)}")
        logger.info(f"Rank B: {np.linalg.matrix_rank(B)}")

        if self.regen_states:
            X = fixpoint_iteration(A, B, U, model_config["activation_fn"], model_config["device"], atol=model_config["atol"]).cpu()

        logger.info("===== Start parallel solve for C and D =====")
        CD = self.state_matching(np.hstack([X.cpu().numpy().T, U.cpu().numpy().T]), Y.T)
        C = CD[:, :n]
        D = CD[:, n:]

        return A, B, C, D

    def state_matching(self, X, Y):
        """
        Solve the least square problem with numpy.linalg.lstsq.
        """
        W, c, r, _ = np.linalg.lstsq(X, Y, rcond=None)

        loss = np.mean(np.square(X @ W - Y))
        
        logger.info(f"Total Lasso loss: {loss}")
        logger.info(f"Data rank: {r}")

        W[np.abs(W) <= self.tol] = 0
        
        return W.T

    def lowrank_solve_matrix(self, X, U, Z, n, p, kappa, device, plot_loss):
        """
        Solve the state matching problem with projected gradient descent to ensure A is low-rank and well-posed.
        """
        model = LowRankModel(n=n, p=p, rank=self.rank, kappa=kappa, device=device)
        trainer = Trainer(num_epoch=self.num_epoch, lambda_z=self.lambda_z, lr=self.lr, verbose_epoch=self.verbose_epoch)

        model, losses = trainer.train(model, X, U, Z)

        # Plot losses 
        if plot_loss:
            plt.figure()
            plt.plot(losses)
            plt.xlabel("Epoch")
            plt.ylabel("MSE loss")
            plt.yscale("log")
            plt.title(f"Training Loss")
            plt.savefig(f"loss_AB.png")

            # Save the loss trace
            np.save(f"loss_AB_trace.npy", losses)

        L = model.L.clone().detach().cpu().numpy()
        R = model.R.clone().detach().cpu().numpy()
        B = model.B.clone().detach().cpu().numpy()

        logger.info(f"Total loss: {losses[-1]}")

        return L, R, B




import gc
import logging
import math
from multiprocessing import Pool, shared_memory
from typing import Any, Dict, Tuple
import cvxpy as cp
import numpy as np
from scipy import sparse
from tqdm import tqdm

from .solver import BaseSolver
from ..utils import fixpoint_iteration

logger = logging.getLogger(__name__)

def create_shared_memory_block(ndarray_to_share):
    """
    Create a shared memory block for the given ndarray to be used by multiple processes.
    """
    # create a shared memory of size array.nbytes
    shm_blocks = [
        shared_memory.SharedMemory(create=True, size=array.nbytes)
        for array in ndarray_to_share
    ]
    # create a ndarray using the buffer of shm
    ndarray_shm = [
        np.ndarray(array.shape, dtype=array.dtype, buffer=shm.buf)
        for (array, shm) in zip(ndarray_to_share, shm_blocks)
    ]
    # copy the data into the shared memory
    for array_shm, array in zip(ndarray_shm, ndarray_to_share):
        array_shm[:] = array[:]
    return shm_blocks, ndarray_shm

class CVXSolver(BaseSolver):
    r"""
    Train State-driven Implicit Model using CVXPY.
    This is highly recommended for small-scale problems due to its high precision and easy tuning.

    Args:
        num_row (int, optional): Number of rows to solve in each process each iteration. Defaults to 1.
        batch_size (int, optional): Number of columns to solve in each solving iteration. Defaults to 32.
        num_processes (int, optional): Number of processes available. Defaults to 32.
        lambda_y (float, optional): Lasso regularization parameter for Y. Defaults to 1e-6.
        lambda_z (float, optional): Lasso regularization parameter for Z. Defaults to 1e-6.
        l1_ratio (float, optional): Elastic Net regularization parameter. Defaults to 0.5.
        elastic_net (bool, optional): Whether to use Elastic Net regularization. Defaults to False.
        regularized (bool, optional): Whether to use L1 regularization. Defaults to True.
        well_pose (bool, optional): Whether to enforce well-posedness constraint. Defaults to True.
        regen_states (bool, optional): Whether to regenerate states. Defaults to False.
        tol (float, optional): Tolerance for zeroing out weights. Defaults to 1e-6.
    """
    def __init__(
        self,
        num_row : int = 1,
        batch_size : int = 32,
        num_processes : int = 32,
        lambda_y : float = 1e-6,
        lambda_z : float = 1e-6,
        l1_ratio : float = 0.5,
        elastic_net : bool = False,
        regularized : bool = True,
        well_pose : bool = True,
        regen_states : bool = False,
        tol : float = 1e-6,
        verbose : bool = False,
    ):
        self.num_row = num_row
        self.batch_size = batch_size
        self.processes = num_processes
        self.lambda_z = lambda_z
        self.lambda_y = lambda_y
        self.l1_ratio = l1_ratio
        self.elastic_net = elastic_net
        self.regularized = regularized
        self.well_pose = well_pose
        self.regen_states = regen_states
        self.tol = tol
        self.verbose = verbose

    def solve(
        self, 
        X : np.ndarray, 
        U : np.ndarray, 
        Z : np.ndarray, 
        Y : np.ndarray, 
        model_config : Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve State-driven Implicit Model using CVXPY.

        Args:
            X (np.ndarray): Post-activation array.
            U (np.ndarray): Input array.
            Z (np.ndarray): Pre-activation array.
            Y (np.ndarray): Output array.
            model_config (Dict[str, Any]): Model configuration.
                - activation_fn (Callable): Activation function used by the implicit model.
                - device (str): Implicit model's device.
                - atol (float): Equilibrium function's tolerance.
                - kappa (float): Wellposedness condition parameter.

        Returns:
            A, B, C, D (np.ndarray): Implicit model's parameters.
        """
        n, m, p, q = X.shape[0], X.shape[1], U.shape[0], Y.shape[0]

        (X_shm, U_shm), (X, U) = create_shared_memory_block([X, U])

        logger.info("===== Start parallel solve for A and B =====")
        A, B = self.parallel_solve_matrix(X_shm, U_shm, Z, False, (n, m, p, q), model_config["kappa"])

        if self.regen_states:
            X = fixpoint_iteration(A, B, U, model_config["activation_fn"], model_config["device"], atol=model_config["atol"]).cpu().numpy()
            (X_shm, U_shm), (X, U) = create_shared_memory_block([X, U])

        logger.info("===== Start parallel solve for C and D =====")
        C, D = self.parallel_solve_matrix(X_shm, U_shm, Y, True, (n, m, p, q), model_config["kappa"])

        X_shm.close()
        X_shm.unlink()
        U_shm.close()
        U_shm.unlink()

        return A, B, C, D

    def parallel_solve_matrix(self, X_shm, U_shm, YZ, is_y, problem_size, kappa):
        """
        Dividing the data matrix and solve parallelly on CPU over multiple processes.
        """
        # initialize empty list to store csr format sparse matrix
        A, B = None, None
        loss = []

        # start batch processing
        total_processes = math.ceil(YZ.shape[0] / self.num_row)
        for batch in tqdm(range(0, YZ.shape[0], self.batch_size)):
            # construct parallel input data for a batch
            batch_end = min(batch + self.batch_size, YZ.shape[0])
            parallel_input = [
                (
                    X_shm.name,
                    U_shm.name,
                    YZ[i : min(i + self.num_row, YZ.shape[0])].T,
                    is_y,
                    problem_size,
                    kappa,
                )
                for i in range(batch, batch_end, self.num_row)
            ]
            # logger.info(f"Solving batch {batch}")

            # construct cvxpy with multiprocessing
            with Pool(processes=self.processes) as pool:
                results = []

                for result in pool.starmap(self.solve_opt_problem, parallel_input):
                    results.append(result)

            # unzip a list of tuples
            A_lst, B_lst, losses = list(zip(*results))
            # store in scipy csr_matrix format
            A_sp, B_sp = sparse.vstack(A_lst), sparse.vstack(B_lst)
            A = sparse.vstack([A, A_sp]) if A is not None else A_sp
            B = sparse.vstack([B, B_sp]) if B is not None else B_sp
            del A_lst, B_lst, A_sp, B_sp  # free memory
            gc.collect()

            loss.extend(losses)

        loss = np.sum(loss)
        logger.info(f"Total loss: {loss}")

        return A, B

    def solve_opt_problem(self, X_shm_name, U_shm_name, yz, is_y, problem_size, kappa):
        """
        Solve the optimization problem using CVXPY.
        """
        n, m, p, _ = problem_size

        X_shm = shared_memory.SharedMemory(name=X_shm_name)
        U_shm = shared_memory.SharedMemory(name=U_shm_name)
        X = np.ndarray((n, m), dtype="float32", buffer=X_shm.buf)
        U = np.ndarray((p, m), dtype="float32", buffer=U_shm.buf)

        # variables for model weights
        a = cp.Variable((n, self.num_row))
        b = cp.Variable((p, self.num_row))

        # set up parameters
        yz_param = cp.Parameter((m, self.num_row))

        objective, constraints = self.get_objective_cvxpy(
            a,
            b,
            yz_param,
            X,
            U,
            is_y,
            kappa,
        )
        yz_param.value = yz
        prob = cp.Problem(objective, constraints)

        try:
            prob.solve(verbose=self.verbose, solver=cp.MOSEK)
        except:
            # prob.solve(verbose=self.verbose, solver=cp.ECOS)
            prob.solve(verbose=self.verbose, solver=cp.CLARABEL)

        # threshold entries to enforce exact zeros and store array in compress sparse row format
        a.value[abs(a.value) <= self.tol] = 0
        b.value[abs(b.value) <= self.tol] = 0
        # logger.info(f"1-norm of a: {np.linalg.norm(a.value, 1)}.")

        if self.elastic_net:
            loss = self.elastic_net_loss(yz, a.value, b.value, X, U, is_y)
            # logger.info(f"ElasticNet loss: {elastic_loss}")
        elif self.regularized:
            loss = self.lasso_loss(yz, a.value, b.value, X, U, is_y)
        else:
            loss = 0

        a_sp = sparse.csr_matrix(a.value.T)
        b_sp = sparse.csr_matrix(b.value.T)

        return a_sp, b_sp, loss

    def get_objective_cvxpy(self, a, b, yz_param, X, U, is_y, kappa):
        """
        Return objective and constraints for sequential solver.
        Note: robust optimizer formula: minimize vectorized L1 min(sim|M_i,j|)
        """
        objective = 0
        constraints = []
        if is_y:
            lambda_yz = self.lambda_y
        else:
            lambda_yz = self.lambda_z

        if self.elastic_net:
            # Elastic Net
            objective += lambda_yz * self.l1_ratio * (cp.pnorm(a, 1) + cp.pnorm(b, 1))
            objective += 0.5 * lambda_yz * (1 - self.l1_ratio) * (cp.pnorm(a, 2)**2 + cp.pnorm(b, 2)**2)
        else:
            # L1 objective
            objective = objective + lambda_yz * (cp.pnorm(a, 1) + cp.pnorm(b, 1))

        if self.regularized:
            objective = (
                objective
                # + 1/(2 * X.shape[1]) * cp.pnorm(yz_param - (np.hstack([X.T, U.T]) @ cp.vstack([a, b])), 2) ** 2
                + 0.5 * cp.pnorm(yz_param - (np.hstack([X.T, U.T]) @ cp.vstack([a, b])), 2) ** 2
            )
        else:
            # exact matching constraint
            constraints.append((X.T @ a) + (U.T @ b) == yz_param)

        # well-posedness constraint
        if not is_y and self.well_pose:
            constraints.append(cp.pnorm(a, 1) <= kappa)

        return cp.Minimize(objective), constraints

    def elastic_net_loss(self, yz, a, b, X, U, is_y):
        """Calculate the Elastic Net loss."""
        if is_y:
            lambda_yz = self.lambda_y
        else:
            lambda_yz = self.lambda_z
        # Calculate L1 regularization term
        l1_term = self.l1_ratio * (np.sum(np.abs(a)) + np.sum(np.abs(b)))

        # Calculate L2 regularization term
        l2_term = (1 - self.l1_ratio) * 0.5 * (np.sum(np.square(a)) + np.sum(np.square(b)))

        # Calculate the mean squared error (MSE)
        mse = np.sum(np.square(yz - (np.hstack([X.T, U.T]) @ np.vstack([a, b]))))

        # Calculate the total loss
        loss = 1/(2 * X.shape[1]) * mse + lambda_yz * (l1_term + l2_term)

        return loss

    def lasso_loss(self, yz, a, b, X, U, is_y):
        """Calculate the Lasso loss."""
        if is_y:
            lambda_yz = self.lambda_y
        else:
            lambda_yz = self.lambda_z
        # Calculate L1 regularization term
        l1_term = np.sum(np.abs(a)) + np.sum(np.abs(b))

        # Calculate the mean squared error (MSE)
        mse = np.sum(np.square(yz - (np.hstack([X.T, U.T]) @ np.vstack([a, b]))))

        # Calculate the total loss
        # loss = 1/(2 * X.shape[1]) * mse + lambda_yz * l1_term
        loss = 0.5 * mse + lambda_yz * l1_term

        return loss




import torch
from torch import nn
import torch.nn.functional as F
from typing import Optional, Type
from .implicit_function import ImplicitFunctionInf, ImplicitFunction, transpose, project_onto_Linf_ball


class ImplicitModel(nn.Module):
    r""" 
    The most basic form of an Implicit Model.
    
    **Note**: In conventional deep learning, the batch size typically comes first for inputs :math:`U`, hidden states :math:`X`, and outputs :math:`Y`. 
    We follow this convention, but the model internally transposes these matrices to solve the fixed-point equation.
    Users can input their data in the usual format, and the output will be returned in the standard format. 

    Args:
        input_dim (int): Number of input features (:math:`p`).
        output_dim (int): Number of output features (:math:`q`).
        hidden_dim (int): Number of hidden features (:math:`n`).
        is_low_rank (bool, optional): Whether to use low-rank approximation (default: False).
        rank (int, optional): Rank for low-rank approximation (:math:`r`), required if `is_low_rank` is True.
        f (Type[ImplicitFunction], optional): The implicit function to use (default: ImplicitFunctionInf for well-posedness).
        kappa (float, optional): Radius of the L-infinity norm ball (:math:`\kappa`) for well-posedness. (default: 0.99).
        no_D (bool, optional): Whether to exclude matrix D (default: False).
        bias (bool, optional): Whether to include a bias term (default: False).
        mitr (int, optional): Max iterations for the forward pass. (default: 300).
        grad_mitr (int, optional): Max iterations for gradient computation. (default: 300).
        tol (float, optional): Convergence tolerance for the forward pass. (default: 3e-6).
        grad_tol (float, optional): Convergence tolerance for gradients. (default: 3e-6).
    """
    
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dim: int,
        f: Type[ImplicitFunction] = ImplicitFunctionInf,
        no_D: bool = False,
        bias: bool = False,
        mitr: int = 300,
        grad_mitr: int = 300,
        tol: float = 3e-6,
        grad_tol: float = 3e-6,
        kappa: float = 0.99,
        is_low_rank: bool = False,
        rank: Optional[int] = None
    ) -> None:
        super().__init__()

        if is_low_rank and rank is None:
            raise ValueError("Parameter 'k' is required when 'is_low_rank' is True.")

        if bias:
            input_dim += 1

        self.hidden_dim: int = hidden_dim
        self.input_dim: int = input_dim
        self.output_dim: int = output_dim
        self.is_low_rank: bool = is_low_rank
        self.no_D: bool = no_D
        self.bias: bool = bias
        
        if self.is_low_rank:
            if rank is None:
                raise ValueError("Rank must be specified when using low-rank approximation.")
            self.L: nn.Parameter = nn.Parameter(torch.randn(hidden_dim, rank) / hidden_dim)
            self.R: nn.Parameter = nn.Parameter(torch.randn(hidden_dim, rank) / hidden_dim)
        else:
            self.A: nn.Parameter = nn.Parameter(torch.randn(hidden_dim, hidden_dim) / hidden_dim)

        self.B: nn.Parameter = nn.Parameter(torch.randn(hidden_dim, input_dim) / hidden_dim)
        self.C: nn.Parameter = nn.Parameter(torch.randn(output_dim, hidden_dim) / hidden_dim)
        if not self.no_D:
            self.D: nn.Parameter = nn.Parameter(torch.randn(output_dim, input_dim) / hidden_dim)
        else:
            self.D: torch.Tensor = torch.zeros((output_dim, input_dim), requires_grad=False)

        self.f: ImplicitFunction = f()
        self.f.set_parameters(mitr=mitr, grad_mitr=grad_mitr, tol=tol, grad_tol=grad_tol, kappa=kappa)


    def forward(
        self, 
        U: torch.Tensor, 
        X0: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward pass of `ImplicitModel`.

        Args:
            U (torch.Tensor): Input tensor of shape (batch_size, input_dim).
            X0 (torch.Tensor, optional): Initial hidden state tensor of shape (batch_size, hidden_dim).

        Returns:
            torch.Tensor: The output tensor of shape (batch_size, output_dim).
        """

        if (len(U.size()) == 3):
            U = U.flatten(1, -1)
        U = transpose(U)
        if self.bias:
            U = F.pad(U, (0, 0, 0, 1), value=1)
        assert U.shape[0] == self.input_dim, f'Given input size {U.shape[0]} does not match expected input size {self.p}.'

        m = U.shape[1]
        X_shape = torch.Size([self.hidden_dim, m])

        if X0 is not None:
            X0 = transpose(X0)
            assert X0.shape == X_shape
        else:
            X0 = torch.zeros(X_shape, dtype=U.dtype, device=U.device)

        if self.is_low_rank:
            L_projected = project_onto_Linf_ball(self.L, self.f.kappa)
            RT_projected = project_onto_Linf_ball(transpose(self.R), self.f.kappa)
            X = self.f.apply(L_projected @ RT_projected, self.B, X0, U)
            
        else:
            X = self.f.apply(self.A, self.B, X0, U)
        return transpose(self.C @ X + self.D @ U)


import numpy as np
import torch
from .utils import set_seed

set_seed(0)

# Spiky region function
def spiky_function(x, scale=5, freq=[2, 23, 78, 100]):
    return scale * sum(np.sin(f * x) for f in freq)

# Non-spiky region function with added noise
def non_spiky_function(x, noise_mean=0, noise_std=0.25):
    return np.sin(x) + np.random.normal(noise_mean, noise_std, size=x.shape)

level = 7
total_data_points = 10000
num_spiky_regions = 20 + level
spiky_region_size = 100 + level
spiky_region_magnitude = level/2

# Generate data
x_values = np.linspace(0, 100, total_data_points)
y_values = np.zeros(total_data_points)

# Assign spiky regions
spiky_regions = torch.randint(0, total_data_points - spiky_region_size, (num_spiky_regions, ))
for start_idx in spiky_regions:
    end_idx = start_idx + spiky_region_size
    x_spiky = x_values[start_idx:end_idx]
    y_values[start_idx:end_idx] = spiky_function(x_spiky, spiky_region_magnitude)

# Assign non-spiky regions
mask = y_values == 0
y_values[mask] = non_spiky_function(x_values[mask], noise_mean=0, noise_std=level*0.05)

# Save and reload data
np.savetxt(f'./data/spiky_data_{level}.csv', y_values, delimiter=',', fmt='%.17g')
y_values = np.loadtxt(f'./data/spiky_data_{level}.csv', delimiter=',')


def load_data(stock, look_back):
    data_raw = stock 
    data = []
    
    # create all possible sequences of length look_back
    for index in range(len(data_raw) - look_back): 
        data.append(data_raw[index:index + look_back])
    data = np.array(data)

    test_set_size = int(np.round(0.4*data.shape[0])) # 30% for test
    train_set_size = data.shape[0] - (test_set_size)
    
    x_train = data[:train_set_size,:-1,:]
    y_train = data[:train_set_size,-1,:]
    x_test = data[train_set_size:,:-1]
    y_test = data[train_set_size:,-1,:]

    x_train = torch.from_numpy(x_train).type(torch.Tensor)
    x_test = torch.from_numpy(x_test).type(torch.Tensor)
    y_train = torch.from_numpy(y_train).type(torch.Tensor)
    y_test = torch.from_numpy(y_test).type(torch.Tensor)

    return x_train, y_train, x_test, y_test

def spiky_synthetic_dataset(look_back):
    data = y_values.reshape(-1,1)
    x_train, y_train, x_test, y_test = load_data(data, look_back)
    return x_train, x_test, y_train, y_test


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
from datetime import datetime

# Add the project's source directory to the Python path
sys.path.insert(0, os.path.abspath('../../src'))

# -- Project information -----------------------------------------------------

project = 'Implicit Deep Learning'
copyright = f'2025, Hoang Phan, Bao Tran, Chi Nguyen, Bao Truong, Thanh Tran, Khai Nguyen, Hong Chu, Alicia Y. Tsai, Laurent El Ghaoui'
author = 'Hoang Phan, Bao Tran, Chi Nguyen, Bao Truong, Thanh Tran, Khai Nguyen, Hong Chu, Alicia Y. Tsai, Laurent El Ghaoui'
release = '0.0.2'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinxemoji.sphinxemoji',  # For emoji support
    'sphinx_copybutton',  # For copy button in code blocks
    'sphinx_markdown_tables',  # For Markdown table support
    'sphinx_autodoc_typehints',
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
napoleon_use_ivar = True

# Intersphinx mapping
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'torch': ('https://pytorch.org/docs/stable', None),
    'numpy': ('https://numpy.org/doc/stable', None),
}

templates_path = ['_templates']
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"
pygments_dark_style = "gruvbox-dark"
highlight_language = "python3"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = 'furo' 

# Theme options
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/HoangP8/Implicit-Deep-Learning",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    "source_repository": "https://github.com/HoangP8/Implicit-Deep-Learning/",
    "source_branch": "main",
    "source_directory": "docs/source/",
}

html_static_path = ['_static']
# html_css_files = ['css/custom.css']


"""

This file contains a simple example code for training an explicit network, which is later used to generate states to train the Implicit Model.

"""

import os
import logging
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
from torch.utils.data import DataLoader
from torchvision import transforms, datasets

from explicit_networks import FashionMNIST_FFNN

logger = logging.getLogger(__name__)

def load_data(data_dir="data", batch_size=128, num_workers=4):
    transform = transforms.Compose([transforms.ToTensor()])

    train_set = datasets.FashionMNIST(
        f"{data_dir}/FashionMNIST",
        train=True,
        download=True,
        transform=transform,
    )
    test_set = datasets.FashionMNIST(
        f"{data_dir}/FashionMNIST",
        train=False,
        download=True,
        transform=transform,
    )

    train_loader = DataLoader(
        train_set,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
    )
    test_loader = DataLoader(
        test_set,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
    )
    return train_loader, test_loader


def train(model, train_loader, loss_fn, optimizer, epoch, device):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % 100 == 0:
            loss = loss.item()
            logger.info(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss / len(data),
                )
            )

def test(model, test_loader, loss_fn, epoch, device):
    model.eval()
    test_loss = torch.tensor(0.0, device=device)
    correct = torch.tensor(0.0, device=device)
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += loss_fn(output, target)
            _, preds = torch.max(output, 1)
            correct += torch.sum(preds == target.data).item()

    test_loss = (test_loss / len(test_loader.dataset)).item()
    correct = correct.item()

    logger.info(
        "Test Epoch: {} Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n".format(
            epoch,
            test_loss,
            correct,
            len(test_loader.dataset),
            100.0 * correct / len(test_loader.dataset),
        )
    )


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--batch_size", type=int, default=128)
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--data_dir", type=str, default="data")
    parser.add_argument("--learning_rate", type=float, default=0.01)
    parser.add_argument("--output_dir", type=str, default="models")
    args = parser.parse_args()

    device = "cuda" if torch.cuda.is_available() else "cpu"

    os.makedirs(args.output_dir, exist_ok=True)

    model = FashionMNIST_FFNN(28 * 28, 10)
    
    model = model.to(device)

    num_params = sum(p.numel() for p in model.parameters())
    logger.info(f"Number of parameters: {num_params}")

    train_loader, test_loader = load_data(data_dir=args.data_dir, batch_size=args.batch_size)

    loss_fn = nn.CrossEntropyLoss(reduction="sum")

    train_params = [param for param in model.parameters()]
    logger.info(f"Number of trainable parameters: {sum(p.numel() for p in train_params)}")

    optimizer = optim.SGD(train_params, lr=args.learning_rate, momentum=0.9, weight_decay=0.0001)

    scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=args.epochs, eta_min=0.0001)

    for epoch in tqdm(range(args.epochs)):
        train(model, train_loader, loss_fn, optimizer, epoch, device)
        test(model, test_loader, loss_fn, epoch, device)
        scheduler.step()

    torch.save(model.state_dict(), f"{args.output_dir}/explicit_model.pt")



from abc import ABC, abstractmethod
from typing import Dict, Any, Tuple
import numpy as np

class BaseSolver(ABC):
    r"""
    Base class for all solvers. All solver implementations must inherit from this class
    and implement the solve method.
    """
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def solve(
        self, 
        X: np.ndarray, 
        U: np.ndarray, 
        Z: np.ndarray, 
        Y: np.ndarray, 
        model_config: Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Solve for the implicit model parameters.

        Args:
            X (np.ndarray): Post-activation array of shape (n_samples, hidden_dim).
            U (np.ndarray): Input array of shape (n_samples, input_dim).
            Z (np.ndarray): Pre-activation array of shape (n_samples, hidden_dim).
            Y (np.ndarray): Output array of shape (n_samples, output_dim).
            model_config (Dict[str, Any]): Model configuration containing:
                - activation_fn (Callable): Activation function used by the implicit model
                - device (str): Device to run computations on ('cpu' or 'cuda')
                - atol (float): Absolute tolerance for convergence
                - kappa (float): Wellposedness condition parameter

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: A tuple containing:
                - A (np.ndarray): Hidden-to-hidden weight matrix of shape (hidden_dim, hidden_dim)
                - B (np.ndarray): Input-to-hidden weight matrix of shape (hidden_dim, input_dim)
                - C (np.ndarray): Hidden-to-output weight matrix of shape (output_dim, hidden_dim)
                - D (np.ndarray): Input-to-output weight matrix of shape (output_dim, input_dim)
        """
        raise NotImplementedError


"""

This file contains the example code for training a simple Implicit Model with SIM.

"""

import torch
import random
from torch.utils.data import DataLoader, Subset

from idl.sim import SIM
from idl.sim.solvers import ADMMSolver

from explicit_networks import FashionMNIST_FFNN
from train_explicit import load_data


train_loader, test_loader = load_data(data_dir="data")

# Take only a subset of the training dataset to train the state-driven model
selected_indices = random.sample(
    range(len(train_loader.dataset)), 2000
)
subset = Subset(train_loader.dataset, selected_indices)
subset_loader = DataLoader(subset, batch_size=1000, shuffle=True)

explict_model = FashionMNIST_FFNN(28 * 28, 10)
explict_model.load_state_dict(torch.load("models/explicit_model.pt"))

sim = SIM(activation_fn=torch.nn.functional.relu, device="cuda", dtype=torch.float32)

solver = ADMMSolver(
    num_epoch_ab=1500,
    num_epoch_cd=120,
    lambda_y=1.0,
    lambda_z=1.0,
    rho_ab=1.0,
    rho_cd=1.0,
    batch_feature_size=120,
    regen_states=False,
)

# Train SIM
sim.train(solver=solver, model=explict_model, dataloader=subset_loader)

# Evaluate SIM
test_acc = sim.evaluate(test_loader)
print(f"Test accuracy: {test_acc}")



import torch
import torch.nn as nn
import numpy as np
from sklearn.preprocessing import StandardScaler

from contextlib import contextmanager
import logging
from typing import Callable, Dict, List, Optional, Union

from .utils import fixpoint_iteration

logger = logging.getLogger(__name__)

__activation_types__ = (
    nn.CELU,
    nn.ELU,
    nn.GELU,
    nn.GLU,
    nn.Hardshrink,
    nn.Hardsigmoid,
    nn.Hardswish,
    nn.Hardtanh,
    nn.LeakyReLU,
    nn.LogSigmoid,
    nn.LogSoftmax,
    nn.Mish,
    nn.MultiheadAttention,
    nn.PReLU,
    nn.ReLU,
    nn.ReLU6,
    nn.RReLU,
    nn.SELU,
    nn.Sigmoid,
    nn.SiLU,
    nn.Softmax,
    nn.Softmax2d,
    nn.Softmin,
    nn.Softplus,
    nn.Softshrink,
    nn.Softsign,
    nn.Tanh,
    nn.Tanhshrink,
    nn.Threshold,
)

class HookManager:
    def __init__(self):
        """
        HookManager class to collect pre- and post-activations from a model.
        """
        self.pre_activations = []
        self.post_activations = []
        self.hooks = []

    def hook_fn(self, module : torch.nn.Module, input : torch.Tensor, output : torch.Tensor):
        """
        Hook function to collect pre- and post-activations from a model.
        """
        if input[0].size() == output.size():
            self.pre_activations.append(input[0].detach())
            self.post_activations.append(output.detach())

    def _apply_all_hooks(self, module : torch.nn.Module):
        """
        Recursively apply hooks to all activation layers in the model.
        """
        children = list(module.children())
        if len(children) == 0:  # It's a leaf module
            if isinstance(module, __activation_types__): # check if the module is an activation layer
                hook = module.register_forward_hook(self.hook_fn)
                self.hooks.append(hook)
        else:
            for child in module.children():
                self._apply_all_hooks(child)  # Recursively apply to children
    
    def _check_modules(self, module : torch.nn.Module, list_relu : List[torch.nn.Module]):
        """
        Recursively look for activation layers inside the module.
        """
        children = list(module.children())
        if len(children) == 0:  # It's a leaf module
            if isinstance(module, __activation_types__):
                list_relu.append(module)
        else:
            for child in module.children():
                self._check_modules(child, list_relu)  # Recursively apply to children

    def _apply_n_hooks(self, module : torch.nn.Module, n : int = 3):
        """
        Skip some activations and apply hooks only to the last activation layer of each layer block.

        Args:
            module (torch.nn.Module): The module to apply hooks to.
            n (int): The threshold to define a block of layers. Only skip layers when the number of children is greater than n.
        """
        children_length = len(list(module.children()))
        if children_length == 0:
            if isinstance(module, __activation_types__):
                hook = module.register_forward_hook(self.hook_fn)
                self.hooks.append(hook)
        elif children_length <= n: # if the number of children is less than n, apply hooks to all children
            for child in module.children():
                self._apply_n_hooks(child, n)
        else: # if the number of children is greater than n, apply hooks to the last activation of the layer block
            for child in module.children():
                relu_layers = []
                self._check_modules(child, relu_layers)
                if relu_layers:
                    final_relu = relu_layers[-1] # get only the last activation of a block of layers
                    hook = final_relu.register_forward_hook(self.hook_fn)
                    self.hooks.append(hook)

    @contextmanager
    def register_hooks(self, model : torch.nn.Module, skip_layers : Optional[int] = None):
        """
        Register hooks to all activation layers in the model.
        """
        if skip_layers is None:
            self._apply_all_hooks(model)
        else:
            self._apply_n_hooks(model, skip_layers)
        try:
            yield
        finally:
            for hook in self.hooks:
                hook.remove()
            self.pre_activations.clear()
            self.post_activations.clear()
            self.hooks.clear()


class SIM():
    r"""
    SIM base class.

    Args:
        activation_fn (Callable): Activation function used in the implicit model.
        kappa (float): Parameter to ensure convergence of the fixed-point iteration. Default is 0.99.
        atol (float): Absolute tolerance for the fixed-point iteration. Default is 1e-6.
        skip_layers (int, optional): If not None, only the last activation of each layer block will be used. The block size is controlled by the parameter. Defaults to None.
        standardize (bool, optional): Whether to standardize the input data using scipy StandardScaler. Defaults to False.
        device (str or torch.device, optional): Device to use for SIM. Defaults to "cpu".
        dtype (str or torch.dtype, optional): Data type for SIM. Defaults to torch.float32.
    """
    def __init__(
        self,
        activation_fn : Callable = nn.functional.relu,
        kappa : float = 0.99,
        atol : float = 1e-6,
        skip_layers : Optional[int] = None,
        standardize : bool = False,
        device : Optional[Union[str, torch.device]] = "cpu", 
        dtype : Optional[Union[str, torch.dtype]] = torch.float32,
    ) -> None:
        self.activation_fn = activation_fn
        self.atol = atol
        self.kappa = kappa
        self.skip_layers = skip_layers
        self.device = device
        self.dtype = dtype
        self.weights = {
            'A': None,
            'B': None,
            'C': None,
            'D': None,
        }
        self.standardize = standardize
        if standardize:
            self.scaler = StandardScaler()

    def to(
        self,
        device : Optional[Union[str, torch.device]] = None, 
        dtype : Optional[Union[str, torch.dtype]] = None
    ) -> None:
        if device is not None:
            self.device = device
        if dtype is not None:
            self.dtype = dtype
        for weight in self.weights.keys():
            if self.weights[weight] is not None:
                self.weights[weight] = self.weights[weight].to(self.device, self.dtype)
        return self
    
    def __call__(
        self, 
        input : torch.Tensor,
    ) -> torch.Tensor:
        """
        Forward pass of a standard implicit model.

        Args:
            input (torch.Tensor): Input data with shape (batch_size, input_dim).
        
        Returns:
            output (torch.Tensor): Output data with shape (batch_size, output_dim).
        """
        for weight in self.weights.keys():
            assert self.weights[weight] is not None, f"Weight matrix {weight} is not trained"

        X = fixpoint_iteration(self.weights['A'], self.weights['B'], input, self.activation_fn, self.device, atol=self.atol)
        Y = X @ self.weights['C'] + self.weights['D'] @ input

        return Y
    
    def get_states(
        self, 
        model : torch.nn.Module, 
        dataloader : torch.utils.data.DataLoader,
    ) -> Dict[str, np.ndarray]:
        """
        Extract the states data (pre-activations, post-activations, inputs, outputs) from the explicit model.
        The dataloader should only load a small amount of data to avoid memory issues.

        Args:
            model (torch.nn.Module): Explicit model (teacher) to extract state data.
            dataloader (torch.utils.data.DataLoader): Data loader.
        
        Returns:
            states_data (dict): Dictionary containing the states data:
                - U (np.ndarray): Input data with shape (batch_size, input_dim).
                - Z (np.ndarray): Pre-activations with shape (batch_size, hidden_dim).
                - X (np.ndarray): Post-activations with shape (batch_size, hidden_dim).
                - Y (np.ndarray): Output data with shape (batch_size, output_dim).
        """
        model.requires_grad_(False)
        model.eval()
        model.to(self.device)
        
        hooks = HookManager()

        outputs_accumulated = []
        inputs_accumulated = []
        pre_activations_accumulated = []
        post_activations_accumulated = []

        for input_samples, _ in dataloader:
            input_samples = input_samples.to(self.device)

            # Run the model with hooks and no gradient calculations
            with hooks.register_hooks(model, skip_layers=self.skip_layers), torch.no_grad():
                outputs = model(input_samples)

                # Accumulate outputs
                outputs_accumulated.append(
                    outputs.cpu().numpy()
                )  # Convert to NumPy array and store

                # Construct and accumulate implicit representation
                U = input_samples.flatten(1).cpu().numpy()
                inputs_accumulated.append(U)

                # Accumulate activations
                num_layers = len(hooks.pre_activations)

                Z = np.hstack(
                    [
                        hooks.pre_activations[i].flatten(1).cpu().numpy()
                        for i in range(num_layers - 1, -1, -1)
                    ]
                )
                pre_activations_accumulated.append(Z)

                X = np.hstack(
                    [
                        hooks.post_activations[i].flatten(1).cpu().numpy()
                        for i in range(num_layers - 1, -1, -1)
                    ]
                )
                post_activations_accumulated.append(X)

        states_data = {}
        states_data['Y'] = np.vstack(outputs_accumulated).T
        states_data['U'] = np.vstack(inputs_accumulated).T
        states_data['Z'] = np.vstack(pre_activations_accumulated).T
        states_data['X'] = np.vstack(post_activations_accumulated).T

        if self.standardize:
            states_data['U'] = self.scaler.fit_transform(states_data['U'].T).T
            states_data['Z'] = self.scaler.fit_transform(states_data['Z'].T).T
            states_data['X'] = self.scaler.fit_transform(states_data['X'].T).T
            states_data['Y'] = self.scaler.fit_transform(states_data['Y'].T).T

        logger.info("===== State Extraction Finished =====")
        logger.info(f"Input data shape: {states_data['U'].shape}")
        logger.info(f"Pre-activation data shape: {states_data['Z'].shape}")
        logger.info(f"Post-activation data shape: {states_data['X'].shape}")
        logger.info(f"Output data shape: {states_data['Y'].shape}")
        return states_data
    
    def train(
        self,
        solver : Callable,
        model : Optional[torch.nn.Module],
        dataloader : Optional[torch.utils.data.DataLoader] = None,
        states_data_path : Optional[str] = None,
        save_states_path : Optional[str] = None,
    ):
        """
        Train the SIM model.

        Args:
            solver (Callable): Solver to use for training.
            model (torch.nn.Module): Explicit model (teacher) to extract state data.
            dataloader (torch.utils.data.DataLoader, optional): Training data loader.
            states_data_path (str, optional): Path to the states data file.
            save_states_path (str, optional): Path to save the states data file.
        """
        # check function arguments for states extraction
        if model is None and states_data_path is None:
            raise ValueError("Either model or states_data_path must be provided")
        elif model is not None and dataloader is None:
            raise ValueError("dataloader must be provided for states extraction")

        if states_data_path is not None and save_states_path is not None:
            logger.warning("Both states_data_path and save_states_path are provided. Only states_data_path will be used.")

        if states_data_path is None:
            logger.info("===== Collecting state matrices from the explicit model =====")
            states_data = self.get_states(model, dataloader)
            if save_states_path is not None:
                logger.info(f"===== Saving states to {save_states_path} =====")
                np.save(save_states_path, states_data)
        else:
            logger.info(f"===== Loading states from {states_data_path} =====")
            states_data = np.load(states_data_path)

        logger.info("===== Start training SIM =====")

        training_config = {
            'activation_fn': self.activation_fn,
            'device': self.device,
            'atol': self.atol,
            'kappa': self.kappa,
        }

        A, B, C, D = solver.solve(states_data['X'], states_data['U'], states_data['Z'], states_data['Y'], training_config)

        logger.info(f"===== Training finished =====")

        self.weights['A'] = A
        self.weights['B'] = B
        self.weights['C'] = C
        self.weights['D'] = D

        return self
    
    def evaluate(
        self,
        dataloader : torch.utils.data.DataLoader,
    ) -> float:
        """
        Evaluate the SIM model on the given test data.

        Args:
            dataloader (torch.utils.data.DataLoader): Test data loader.
        
        Returns:
            test_accuracy (float): Accuracy of the SIM model on the given test data.
        """
        U_test = dataloader.dataset.data.flatten(1).T

        if self.standardize:
            U_test = self.scaler.transform(U_test.T).T 

        X_test = fixpoint_iteration(self.weights['A'], self.weights['B'], U_test, self.activation_fn, self.device).cpu().numpy()
        Y_test_pred = self.weights['C'] @ X_test + self.weights['D'] @ U_test.cpu().numpy()
        test_accuracy = np.mean(
            np.argmax(Y_test_pred, axis=0) == dataloader.dataset.targets.cpu().numpy()
        )
        print(f"Test accuruacy: {(test_accuracy * 100):.2f}%")
        return test_accuracy



from .admm_consensus_multigpu import ADMMMultiGPUSolver
from .admm_consensus import ADMMSolver
from .cvx import CVXSolver
from .projected_gd_lowrank import ProjectedGDLowRankSolver
from .least_square import LeastSquareSolver

__all__ = [
    "ADMMSolver",
    "ADMMMultiGPUSolver",
    "CVXSolver",
    "ProjectedGDLowRankSolver",
    "LeastSquareSolver",
]


import torch
import datetime
import os
import random
import numpy as np

def transpose(X):
    """
    Convenient function to transpose a matrix.
    """
    assert len(X.size()) == 2, "data must be 2D"
    return X.T

def set_seed(seed: int) -> None:
    """
    Set seed for reproducibility.
    """
    random.seed(seed) 
    np.random.seed(seed) 
    torch.manual_seed(seed) 
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

def get_test_accuracy(model, loss_fn, test_loader, device):
    """
    Run the model with the test dataset and compute the loss and accuracy.
    """
    model.eval() # eval mode
    total_loss = 0.0
    correct_predictions = 0
    total_samples = 0

    with torch.no_grad(): 
        for inputs, targets in test_loader:
            inputs, targets = inputs.to(device), targets.to(device)
            predictions = model(inputs)
            if not isinstance(predictions, torch.Tensor):
                predictions = torch.from_numpy(predictions).to(device)

            # Compute loss
            loss = loss_fn(predictions, targets)
            total_loss += loss.item()

            # Compute accuracy
            predicted_labels = torch.argmax(predictions, dim=-1)
            correct_predictions += (predicted_labels == targets).sum().item()
            total_samples += targets.size(0)

    avg_loss = total_loss / len(test_loader)
    accuracy = correct_predictions / total_samples
    model.train() # train mode
    return avg_loss, accuracy


def train(args, model, train_loader, test_loader, optimizer, loss_fn, num_epochs, log_dir, device):
    """
    Trains the model for a specified number of epochs and logs the results.

    Args:
        model (torch.nn.Module): The model to train.
        train_loader (torch.utils.data.DataLoader): The training data loader.
        test_loader (torch.utils.data.DataLoader): The test data loader.
        optimizer (torch.optim.Optimizer): The optimizer used for training.
        loss_fn (torch.nn.Module): The loss function to minimize.
        num_epochs (int): The number of epochs to train the model.
        log_dir (str): The base directory to save log files.
        device (torch.device, optional): The device (CPU or GPU) to perform training on.

    Returns:
        tuple: The trained model and the log file path.
    """

    model.to(device)

    # Prepare logging
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")    
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_file_path = os.path.join(log_dir, f"{timestamp}.log")
    log_file = open(log_file_path, 'w')
    
    log_file.write("Arguments:\n")
    for arg, value in vars(args).items():
        log_file.write(f"{arg}: {value}\n")
    log_file.write("\n")
    log_file.write(f'Model size: {sum(p.numel() for p in model.parameters())} parameters\n')
    log_file.write("\n")
    log_file.write("Epoch,Train Loss,Train Accuracy,Test Loss,Test Accuracy\n")

    # Training loop
    for epoch in range(num_epochs):
        model.train()  # train mode
        train_loss = 0.0
        correct_train = 0
        total_train = 0

        batch_idx = 0
        total_batches = len(train_loader) 

        for inputs, targets in train_loader:
            inputs, targets = inputs.to(device), targets.to(device)

            # Forward pass
            predictions = model(inputs)
            loss = loss_fn(predictions, targets)

            # Backward pass
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            predicted_labels = torch.argmax(predictions, dim=-1)
            correct_train += (predicted_labels == targets).sum().item()
            total_train += targets.size(0)
            train_loss += loss.item()

            if batch_idx % 100 == 0:
                test_loss, test_accuracy = get_test_accuracy(model, loss_fn, test_loader, device)

                print(f"Epoch {epoch + 1}/{num_epochs} - "
                      f"Batch {batch_idx + 1}/{total_batches}: "
                      f"Train Loss: {loss.item():.4f}, Train Accuracy: {100 * correct_train / total_train:.2f}%, "
                      f"Test Loss: {test_loss:.4f}, Test Accuracy: {test_accuracy * 100:.2f}%")

            
                log_file.write(f"Epoch {epoch + 1}/{num_epochs} - "
                        f"Batch {batch_idx + 1}/{total_batches}: "
                        f"Train Loss: {loss.item():.4f}, Train Accuracy: {100 * correct_train / total_train:.2f}%, "
                        f"Test Loss: {test_loss:.4f}, Test Accuracy: {test_accuracy * 100:.2f}%\n")

            batch_idx += 1

        avg_train_loss = train_loss / len(train_loader)
        train_accuracy = (correct_train / total_train) * 100
        test_loss, test_accuracy = get_test_accuracy(model, loss_fn, test_loader, device)

        print(f"Epoch {epoch + 1}/{num_epochs}: "
              f"Train Loss: {avg_train_loss:.4f} | Train Accuracy: {train_accuracy:.2f}% | "
              f"Test Loss: {test_loss:.4f} | Test Accuracy: {test_accuracy * 100:.2f}%")
        log_file.write(f"Epoch {epoch + 1}/{num_epochs}: "
                f"Train Loss: {avg_train_loss:.4f} | Train Accuracy: {train_accuracy:.2f}% | "
                f"Test Loss: {test_loss:.4f} | Test Accuracy: {test_accuracy * 100:.2f}%\n")

        log_file.write("-" * 100 + "\n")
        print("-" * 100)

    return model, log_file


import gc
import logging
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
import os
from typing import Any, Dict, List, Optional, Union, Tuple

from .solver import BaseSolver
from ..utils import fixpoint_iteration

logger = logging.getLogger(__name__)


class ADMMMultiGPUSolver(BaseSolver):
    r"""
    ADMM Consensus Solver with distributed computation over multiple GPUs.
    This is highly recommended for large-scale problems and when multiple GPUs are available.

    Args:
        gpu_ids (List[Union[int, torch.device]]): List of GPU IDs or devices.
        num_epoch_ab (int, optional): Number of epochs for solving A and B. Defaults to 1000.
        num_epoch_cd (int, optional): Number of epochs for solving C and D. Defaults to 100.
        lambda_y (float, optional): Lasso regularization parameter for Y. Defaults to 1.0.
        lambda_z (float, optional): Lasso regularization parameter for Z. Defaults to 1.0.
        rho_ab (float, optional): ADMM's rho parameter for A and B. Defaults to 10.0.
        rho_cd (float, optional): ADMM's rho parameter for C and D. Defaults to 10.0.
        batch_feature_size (int, optional): Number of columns to solve in each solving iteration. This is used to control the memory usage.
                The solver is performed (total_rows // batch_feature_size + 1) times. Defaults to 100.
        regen_states (bool, optional): Whether to regenerate states. Defaults to False.
        tol (float, optional): Tolerance for zeroing out weights. Defaults to 1e-6.
    """
    def __init__(
        self,
        gpu_ids : List[Union[int, torch.device]],
        num_epoch_ab : int = 1000,
        num_epoch_cd : int = 100,
        lambda_y : float = 1e-6,
        lambda_z : float = 1e-6,
        rho_ab : float = 10.0,
        rho_cd : float = 10.0,
        batch_feature_size : int = 100,
        regen_states : bool = False,
        tol : float = 1e-6,
    ):
        super().__init__()
        self.gpu_ids = gpu_ids
        self.num_epoch_ab = num_epoch_ab
        self.num_epoch_cd = num_epoch_cd
        self.lambda_y = lambda_y
        self.lambda_z = lambda_z
        self.rho_ab = rho_ab
        self.rho_cd = rho_cd
        self.batch_feature_size = batch_feature_size
        self.regen_states = regen_states
        self.tol = tol

    def solve(
        self, 
        X : np.ndarray, 
        U : np.ndarray, 
        Z : np.ndarray, 
        Y : np.ndarray,
        model_config : Dict[str, Any],
        plot_loss : Optional[bool] = False,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Solve State-driven Implicit Model using the ADMM consensus algorithm.

        Args:
            X (np.ndarray): Post-activation array.
            U (np.ndarray): Input array.
            Z (np.ndarray): Pre-activation array.
            Y (np.ndarray): Output array.
            model_config (Dict[str, Any]): Model configuration.
                - activation_fn (Callable): Activation function used by the implicit model.
                - device (str): Implicit model's device.
                - atol (float): Equilibrium function's tolerance.
                - kappa (float): Wellposedness condition parameter.
            plot_loss (Optional[bool], optional): Whether to plot loss curve. Useful for tuning hyperparameters and debugging. Defaults to False.

        Returns:
            A, B, C, D (np.ndarray): Implicit model's parameters.
        """
        n, m, p, q = X.shape[0], X.shape[1], U.shape[0], Y.shape[0]

        X, U, Z, Y = torch.tensor(X), torch.tensor(U), torch.tensor(Z), torch.tensor(Y)

        logger.info("===== Start solving A and B =====")
        AB = parallel_solve_matrix(torch.hstack([X.T, U.T]), Z.T, is_y=False, n=n, kappa=model_config["kappa"], plot_loss=plot_loss)
        A = AB[:, :n]
        B = AB[:, n:]

        if self.regen_states:
            X = fixpoint_iteration(A, B, U, model_config["activation_fn"], model_config["device"], atol=model_config["atol"]).cpu()

        logger.info("===== Start solving C and D =====")
        CD = parallel_solve_matrix(torch.hstack([X.T, U.T]), Y.T, is_y=True, n=n, kappa=model_config["kappa"], plot_loss=plot_loss)
        C = CD[:, :n]
        D = CD[:, n:]

        return A, B, C, D

    def parallel_solve_matrix(self, X, Y, is_y, n, kappa, plot_loss):
        """
        Dividing the data matrix and distribute batches for parallel solving on multiple GPUs.
        """
        world_size = len(self.gpu_ids)
        total_rows = Y.shape[1]
        batch_rows_length = self.batch_feature_size
        num_batches = total_rows // (batch_rows_length * world_size) + 1

        W = None
        loss = 0.0
        for k in range(num_batches):

            logger.info(f"Solving batch feature {k+1}/{num_batches}")

            start_idx = k * batch_rows_length * world_size
            end_idx = min((k + 1) * batch_rows_length * world_size, total_rows)
            Y_batch = Y[:, start_idx:end_idx]

            # Multiprocessing: spawn multiple processes with torch.multiprocessing
            manager = mp.Manager()
            return_dict = manager.dict()

            mp.set_start_method('spawn', force=True)
            processes = []
            for rank in range(world_size):
                p = mp.Process(target=run_solve_opt_problem, args=(rank, world_size, X, Y_batch, batch_rows_length, is_y, n, kappa, plot_loss, return_dict))
                processes.append(p)
                p.start()
            for p in processes:
                p.join()
            
            # Aggregate results from all processes
            results = [return_dict[i][0] for i in range(world_size)]
            W_k = np.vstack(results)
            W = np.vstack([W, W_k]) if W is not None else W_k

            loss += np.sum([return_dict[i][1] for i in range(world_size)])

            del W_k, results, return_dict
            gc.collect()
        
        logger.info(f"Total Lasso loss: {loss}")
            
        return W

    def setup_process(rank, world_size, backend='nccl'):
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '3090'
        dist.init_process_group(backend, rank=rank, world_size=world_size)

    def cleanup():
        dist.destroy_process_group()

    def run_solve_opt_problem(rank, world_size, X, Y, batch_rows_length, is_y, n, kappa, plot_loss, return_dict):
        """
        ADMM Solve Wrapper.
        """
        setup_process(rank, world_size)
        
        # Set the correct device for this process
        gpu_id = self.gpu_ids[rank]
        torch.cuda.set_device(gpu_id)
        if isinstance(gpu_id, int):
            device = torch.device(f"cuda:{gpu_id}")
        else:
            device = gpu_id

        if is_y:
            num_epoch = self.num_epoch_cd
            rho = self.rho_cd
            lambda_yz = self.lambda_y
        else:
            num_epoch = self.num_epoch_ab
            rho = self.rho_ab
            lambda_yz = self.lambda_z

        Y = Y[:,rank*batch_rows_length:(rank+1)*batch_rows_length]

        if is_y:
            admm = ADMM_CD(X.shape[1], Y.shape[1], rho, lambda_yz, device=device)
        else:
            admm = ADMM_AB(X.shape[1], Y.shape[1], n, rho, lambda_yz, kappa, device=device)

        losses = []
        progress_bar = tqdm(
            range(num_epoch),
            desc="Training Epochs",
            disable=True if rank != 0 else False,
        )
        with torch.no_grad():
            for i in progress_bar:
                admm.step(X, Y)
                loss = admm.LassoObjective(X, Y)
                losses.append(loss)
                progress_bar.update(1)
                progress_bar.set_postfix({"Loss": loss})

        # Plot losses 
        if plot_loss:
            plt.figure()
            plt.plot(losses)
            plt.xlabel("Epoch")
            plt.ylabel("Lasso Objective")
            plt.yscale("log")
            plt.title(f"{rank} Training Loss")
            plt.savefig(f"loss_{rank}_isy_{is_y}.png")

            # Save the loss trace
            np.save(f"loss_trace_{rank}_isy_{is_y}.npy", losses)

        if is_y:
            result = admm.X.T.clone().detach().cpu().numpy()
        else:
            result = admm.avg.T.clone().detach().cpu().numpy()

        result[np.abs(result) <= config.sim.tol] = 0

        # Store the result in the shared dictionary
        return_dict[rank] = (result, losses[-1])
        
        cleanup()


class ADMM_AB:
    def __init__(
        self, 
        D : int, 
        Q : int, 
        n : int, 
        rho : float, 
        lambda_yz : float, 
        kappa : float, 
        device : str
    ):
        """
        Solver class to solve A and B (concatenated) using consensus ADMM.

        Args:
            D (int): Number of features.
            Q (int): Number of samples.
            n (int): Number of rows of A.
            rho (float): ADMM's rho parameter.
            lambda_yz (float): Lasso regularization parameter.
            kappa (float): Wellposedness condition parameter.
            device (str): Device.
        """
        self.D = D
        self.Q = Q
        self.n = n
        self.device = device
        
        self.nu_X = torch.zeros(self.D, self.Q, device=device, requires_grad=False)
        self.nu_Z = torch.zeros(self.D, self.Q, device=device, requires_grad=False)
        self.nu_M = torch.zeros(self.D, self.Q, device=device, requires_grad=False)

        self.rho = rho

        self.X = torch.randn(self.D, self.Q, device=device, requires_grad=False)
        self.Z = torch.zeros(self.D, self.Q, device=device, requires_grad=False)
        self.M = torch.zeros(self.D, self.Q, device=device, requires_grad=False)
        self.avg = torch.zeros(self.D, self.Q, device=device, requires_grad=False)

        self.lambda_yz = lambda_yz
        self.kappa = kappa

    @torch.no_grad()
    def step(self, A, b):
        """
        ADMM's update step.
        """
        A = A.to(self.device)
        b = b.to(self.device)

        t1 = A.T.matmul(A) + self.rho * torch.eye(self.D, device=self.device)
        t2 = A.T.matmul(b) + self.rho * (self.avg - self.nu_X)
        self.X = torch.linalg.solve(t1, t2)

        self.Z = torch.sign(self.avg - self.nu_Z) * torch.clamp(torch.abs(self.avg - self.nu_Z) - self.lambda_yz / self.rho, min=0)

        self.M[:self.n,:] = self.project_w((self.avg - self.nu_M)[:self.n,:].T).T
        self.M[self.n:,:] = (self.avg - self.nu_M)[self.n:,:]

        self.avg = (self.X + self.Z + self.M) / 3

        self.nu_X = self.nu_X + (self.X - self.avg)
        self.nu_Z = self.nu_Z + (self.Z - self.avg)
        self.nu_M = self.nu_M + (self.M - self.avg)

    def project_w(self, matrix):
        """
        Project the matrix to the L1 ball.
        """
        A_np = matrix.clone().cpu().numpy()
        v = self.kappa
        x = np.abs(A_np).sum(axis=-1)

        for idx in np.where(x > v)[0]:
            a_orig = A_np[idx, :]
            a_sign = np.sign(a_orig)
            a_abs = np.abs(a_orig)
            a = np.sort(a_abs)

            s = np.sum(a) - v
            l = float(len(a))
            for i in range(len(a)):
                if s / l > a[i]:
                    s -= a[i]
                    l -= 1
                else:
                    break
            alpha = s / l if l > 0 else np.max(a_abs)
            a = a_sign * np.maximum(a_abs - alpha, 0)
            # assert np.isclose(np.abs(a).sum(), v)
            A_np[idx, :] = a

        proj = torch.tensor(A_np, dtype=self.X.dtype, device=self.device)

        return proj

    @torch.no_grad()
    def LassoObjective(self, A, b):
        """
        Lasso objective function.
        """
        A = A.to(self.device)
        b = b.to(self.device)
        return (0.5 * torch.norm(A.matmul(self.avg) - b)**2 + self.lambda_yz * torch.sum(torch.abs(self.avg))).item()


class ADMM_CD:
    def __init__(
        self, 
        D : int, 
        Q : int, 
        rho : float, 
        lambda_yz : float, 
        device : str
    ):
        """
        Solver class to solve C and D using consensus ADMM.

        Args:
            D (int): Number of features.
            Q (int): Number of samples.
            rho (float): ADMM's rho parameter.
            lambda_yz (float): Lasso regularization parameter.
            device (str): Device.
        """
        self.D = D
        self.Q = Q
        self.device = device
        
        self.nu = torch.zeros(self.D, self.Q, device=device)
        self.rho = rho
        self.X = torch.randn(self.D, self.Q, device=device)
        self.Z = torch.zeros(self.D, self.Q, device=device)
        self.lambda_yz = lambda_yz

    @torch.no_grad()
    def step(self, A, b):
        """
        ADMM's update step.
        """
        A = A.to(self.device)
        b = b.to(self.device)

        t1 = A.T.matmul(A) + self.rho * torch.eye(self.D, device=self.device)
        t2 = A.T.matmul(b) + self.rho * self.Z - self.nu
        self.X = torch.linalg.solve(t1, t2)

        self.Z = self.X + self.nu / self.rho - (self.lambda_yz / self.rho) * torch.sign(self.Z).to(self.device)
        self.nu = self.nu + self.rho * (self.X - self.Z)

    @torch.no_grad()
    def LassoObjective(self, A, b):
        """
        Lasso objective function.
        """
        A = A.to(self.device)
        b = b.to(self.device)
        return (0.5 * torch.norm(A.matmul(self.X) - b)**2 + self.lambda_yz * torch.sum(torch.abs(self.X))).item()


import os
import datetime
import torch
import random
import numpy as np
from typing import Any, Dict
from argparse import Namespace
from torch import nn, Tensor
from torch.optim import Optimizer
from sklearn.metrics import mean_absolute_percentage_error, r2_score


def set_seed(seed: int) -> None:
    """
    Set seed for reproducibility.
    """
    random.seed(seed) 
    np.random.seed(seed) 
    torch.manual_seed(seed) 
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def train_time_series(
    args: Namespace,
    model: nn.Module,
    x_train: Tensor,
    y_train: Tensor,
    x_test: Tensor,
    y_test: Tensor,
    optimizer: Optimizer,
    loss_fn: nn.Module,
    log_dir: str,
    device: torch.device
) -> Dict[str, float]:
    """
    Trains a model on time series data, logs the training process, and evaluates the model.

    Args:
        args (Namespace): Contains configuration such as the number of epochs.
        model (torch.nn.Module): The model to train.
        x_train (torch.Tensor): Training features.
        y_train (torch.Tensor): Training targets.
        x_test (torch.Tensor): Test features.
        y_test (torch.Tensor): Test targets.
        optimizer (torch.optim.Optimizer): The optimizer used for training.
        loss_fn (torch.nn.Module): The loss function to minimize.
        log_dir (str): The directory to save log files.
        device (torch.device): The device (CPU or GPU) for training.

    Returns:
        Dict[str, float]: A dictionary containing model evaluation metrics.
            - 'r2': R-squared score on the test dataset.
            - 'mape': Mean Absolute Percentage Error on the test dataset.
    """
    
    model.to(device)
    x_train, y_train = x_train.to(device), y_train.to(device)
    x_test = x_test.to(device)

    # Prepare logging
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_file_path = os.path.join(log_dir, f"{timestamp}.log")
    log_file = open(log_file_path, 'w')

    log_file.write("Arguments:\n")
    for arg, value in vars(args).items():
        log_file.write(f"{arg}: {value}\n")
    log_file.write("\n")
    log_file.write(f'Model size: {sum(p.numel() for p in model.parameters())} parameters\n')
    log_file.write("\n")

    # Training loop
    for epoch in range(1, args.epochs + 1):
        model.train()
        y_train_pred = model(x_train)
        loss = loss_fn(y_train_pred, y_train)

        if epoch % 10 == 0:
            log_file.write(f"Epoch {epoch}/{args.epochs}, MSE Loss: {loss.item():.4f}")
            print(f"Epoch {epoch}/{args.epochs}, MSE Loss: {loss.item():.4f}")

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    # Evaluation
    model.eval()
    with torch.no_grad():
        y_test_pred = model(x_test).cpu().detach().numpy()
        y_test = y_test.cpu().detach().numpy()

    mape = mean_absolute_percentage_error(y_test, y_test_pred)
    r2 = r2_score(y_test, y_test_pred)

    log_file.write(f'Test R-square: {r2:.2f}\n')
    log_file.write(f'Test MAPE: {mape:.2f}\n')
    log_file.close()
    print(f'Test R square: {r2:.2f}')
    print(f'Test MAPE: {mape:.2f}')


import torch
from torch.autograd import Function
from typing import Optional, Tuple, Any
import numpy as np
import warnings
import torch

def transpose(X: torch.Tensor) -> torch.Tensor:
    """
    Transpose a 2D matrix.
    """
    
    assert X.dim() == 2, "data must be 2D"
    return X.T


def project_onto_Linf_ball(A: torch.Tensor, kappa: float) -> torch.Tensor:
    """
    Project a matrix onto the L-infty norm ball of radius kappa.

    Args:
        A (torch.Tensor): Input matrix.
        kappa (float): The L-infty norm ball radius.

    Returns:
        torch.Tensor: The scaled matrix.
    """

    norm_inf_A = torch.linalg.matrix_norm(A, ord=float('inf')) 
    if norm_inf_A > kappa:
        A = (kappa / norm_inf_A) * A
    return A


class ImplicitFunctionWarning(RuntimeWarning):
    """
    Warning raised when implicit function iterations do not converge.
    """
    pass

class ImplicitFunction(Function):
    """
    Base class for solving implicit functions of the fixed-point equation `AX + BU = X`.
    """
    
    @classmethod
    def set_parameters(
        cls,
        mitr: Optional[int] = None,
        grad_mitr: Optional[int] = None,
        tol: Optional[float] = None,
        grad_tol: Optional[float] = None,
        kappa: Optional[float] = None
    ) -> None:
        """
        Set parameters for Picard iteration and convergence.

        Args:
            mitr (int): Max iterations for the forward pass.
            grad_mitr (int): Max iterations for gradient computation.
            tol (float): Convergence tolerance for the forward pass.
            grad_tol (float): Convergence tolerance for gradients.
            kappa (float): Radius of the L-infinity norm ball for projection.
        """

        if mitr is not None:
            cls.mitr = mitr
        if grad_mitr is not None:
            cls.grad_mitr = grad_mitr
        if tol is not None:
            cls.tol = tol
        if grad_tol is not None:
            cls.grad_tol = grad_tol
        if kappa is not None:
            cls.kappa = kappa                    
    
    @classmethod
    def forward(
        cls,
        ctx: Any,
        A: torch.Tensor,
        B: torch.Tensor,
        X0: torch.Tensor,
        U: torch.Tensor
    ) -> torch.Tensor:
        """
        Perform the forward pass using Picard iterations to solve `AX + BU = X`.

        Args:
            A (torch.Tensor)
            B (torch.Tensor)
            X0 (torch.Tensor)
            U (torch.Tensor)

        Returns:
            torch.Tensor: The stable solution for X
        """

        with torch.no_grad():
            X, err, status = cls.inn_pred(A, B @ U, X0, cls.mitr, cls.tol)
        ctx.save_for_backward(A, B, X, U)
        if status != "converged":
            warnings.warn(f"Picard iterations did not converge: err={err.item():.4e}, status={status}", ImplicitFunctionWarning)
        return X

    @classmethod
    def backward(
        cls,
        ctx: Any,
        *grad_outputs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Compute gradients using implicit backward propagation for the fixed-point equation.

        Args:
            grad_outputs (tuple): Gradients of the output.

        Returns:
            tuple: 
                - Gradient of A.
                - Gradient of B.
                - Zero gradient for X (since X is an output).
                - Gradient of U.
        """

        A, B, X, U = ctx.saved_tensors

        grad_output = grad_outputs[0]
        assert grad_output.size() == X.size()

        DPhi = cls.dphi(A @ X + B @ U)
        V, err, status = cls.inn_pred_grad(A.T, DPhi * grad_output, DPhi, cls.grad_mitr, cls.grad_tol)
        if status != "converged":
            warnings.warn(f"Gradient iterations did not converge: err={err.item():.4e}, status={status}", ImplicitFunctionWarning)
        
        grad_A = V @ X.T
        grad_B = V @ U.T
        grad_U = B.T @ V

        return grad_A, grad_B, torch.zeros_like(X), grad_U

    @staticmethod
    def phi(X: torch.Tensor) -> torch.Tensor:
        """
        ReLU activation.
        """

        return torch.clamp(X, min=0)

    @staticmethod
    def dphi(X: torch.Tensor) -> torch.Tensor:
        """
        Derivative of ReLU.
        """

        grad = X.new_zeros(X.shape)
        grad[X > 0] = 1
        return grad

    @classmethod
    def inn_pred(
        cls,
        A: torch.Tensor,
        Z: torch.Tensor,
        X: torch.Tensor,
        mitr: int,
        tol: float
    ) -> Tuple[torch.Tensor, float, str]:
        """
        Solve `AX + Z = X` using Picard iterations.

        Args:
            A (torch.Tensor)
            Z (torch.Tensor)
            X (torch.Tensor)
            mitr (int): Max iterations.
            tol (float): Convergence tolerance.

        Returns:
            tuple: Solution, error, status ('converged' or 'max itrs reached').
        """

        err = 0
        status = 'max itrs reached'
        for _ in range(mitr):
            X_new = cls.phi(A @ X + Z)
            err = torch.norm(X_new - X, np.inf)
            if err < tol:
                status = 'converged'
                break
            X = X_new
        return X, err, status

    @staticmethod
    def inn_pred_grad(
        AT: torch.Tensor,
        Z: torch.Tensor,
        DPhi: torch.Tensor,
        mitr: int,
        tol: float
    ) -> Tuple[torch.Tensor, float, str]:
        """
        Compute gradient using backward Picard iterations.

        Args:
            AT (torch.Tensor): Transposed matrix A.
            Z (torch.Tensor)
            mitr (int): Max iterations.
            tol (float): Convergence tolerance.
            DPhi (torch.Tensor): Derivative of activation.

        Returns:
            Tuple[torch.Tensor, float, str]: Gradient V, error, status ('converged' or 'max itrs reached').
        """
    
        X = torch.zeros_like(Z)
        err = 0.0
        status = 'max itrs reached'
        for _ in range(mitr):
            X_new = DPhi * (AT @ X) + Z
            err = torch.norm(X_new - X, np.inf)
            if err < tol:
                status = 'converged'
                break
            X = X_new
        return X, err, status


class ImplicitFunctionTriu(ImplicitFunction):
    """
    Implicit function that constrains A to be upper triangular, ensuring a feed-forward architecture.
    
    Methods:
        forward: Applies upper triangular constraint to A and runs the forward pass.
        backward: Keeps the gradient of A upper triangular.
    """

    @classmethod
    def forward(
        cls,
        ctx: Any,
        A: torch.Tensor,
        B: torch.Tensor,
        X0: torch.Tensor,
        U: torch.Tensor
    ) -> torch.Tensor:
        """
        Apply upper triangular constraint to A and perform the forward pass.

        Args:
            ctx (Any): Context object to save information for backward computation.
            A (torch.Tensor): Matrix A.
            B (torch.Tensor): Matrix B.
            X0 (torch.Tensor): Initial hidden state tensor.
            U (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: The stable solution for X.
        """
        
        A = A.triu_(1)
        return super(ImplicitFunctionTriu, cls).forward(ctx, A, B, X0, U)

    @classmethod
    def backward(
        cls,
        ctx: Any,
        *grad_outputs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Keep the gradient of A upper triangular and perform backward pass.

        Args:
            ctx (Any): Context object containing saved tensors.
            grad_outputs (Tuple[torch.Tensor, ...]): Gradients of the output.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
                - Gradient of A (upper triangular).
                - Gradient of B.
                - Zero gradient for X.
                - Gradient of U.
        """
        
        grad_A, grad_B, grad_X, grad_U = super(ImplicitFunctionTriu, cls).backward(ctx, *grad_outputs)
        return grad_A.triu(1), grad_B, grad_X, grad_U


class ImplicitFunctionInf(ImplicitFunction):
    """
    Implicit function that projects A onto the infinity norm ball, enabling closed-loop feedback.
    
    Methods:
        forward: Projects A onto the infinity norm ball and runs the forward pass.
    """

    @classmethod
    def forward(
        cls,
        ctx: Any,
        A: torch.Tensor,
        B: torch.Tensor,
        X0: torch.Tensor,
        U: torch.Tensor
    ) -> torch.Tensor:
        """
        Project A onto the L-infty norm ball and perform the forward pass.

        Args:
            ctx (Any): Context object to save information for backward computation.
            A (torch.Tensor): Matrix A.
            B (torch.Tensor): Matrix B.
            X0 (torch.Tensor): Initial hidden state tensor.
            U (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: The stable solution for X.
        """
        
        A = project_onto_Linf_ball(A, cls.kappa)
        return super(ImplicitFunctionInf, cls).forward(ctx, A, B, X0, U)


import argparse
import torch
from .load_data import load_data
from .utils import train_time_series, set_seed
import os
from idl import ImplicitRNN


def parse_args():
    parser = argparse.ArgumentParser(description="Train an Implicit RNN model on Netflix or Spiky synthetic dataset")
    
    # Dataset
    parser.add_argument('--dataset', type=str, choices=['spiky'], required=True, help="Dataset to use: 'spiky'")
    parser.add_argument('--look_back', type=int, default=60, help="Look-back period for time series data")
    
    # Optimization parameters
    parser.add_argument('--lr', type=float, default=5e-3, help="Learning rate for the optimizer")
    parser.add_argument('--epochs', type=int, default=100, help="Number of epochs for training")
    parser.add_argument('--device', type=int, default=0, help="Specify the device id (e.g., 0 for cuda:0)")
    parser.add_argument("--seed", type=int, default=0, help="Random seed for reproducibility")
    
    # Implicit RNN parameters
    parser.add_argument('--hidden_dim', type=int, default=None, help="Hidden size of RNN model")
    parser.add_argument('--implicit_hidden_dim', type=int, default=None, help="Hidden size of Implicit layer")    
    parser.add_argument('--input_dim', type=int, default=1, help="Input size")
    parser.add_argument('--output_dim', type=int, default=1, help="Output size")
    parser.add_argument('--mitr', type=int, default=500, help="Max iterations")
    parser.add_argument('--grad_mitr', type=int, default=500, help="Max gradient iterations")
    parser.add_argument('--tol', type=float, default=1e-6, help="Tolerance for convergence")
    parser.add_argument('--grad_tol', type=float, default=1e-6, help="Gradient tolerance for convergence")
    parser.add_argument('--is_low_rank', type=bool, default=False, help="Whether to use low rank approximation (True/False)")
    parser.add_argument('--rank', type=int, default=1, help="Rank for low rank approximation (used only if is_low_rank is True)")
    parser.add_argument('--kappa', type=float, default=0.99, help="Inf ball")
    
    return parser.parse_args()

def main():
    """
    Main function to train the Implicit RNN model.
    """

    args = parse_args()
    set_seed(args.seed)
    x_train, x_test, y_train, y_test = load_data(args)

    if args.hidden_dim is None:
        raise ValueError("Error: 'hidden_size' must be specified for the model.")
    elif args.implicit_hidden_dim is None:
        raise ValueError("Error: 'implicit_hidden_dim' must be specified for the model.")
    
    # Initialize the Implicit RNN model
    model = ImplicitRNN(
        input_dim=args.input_dim,
        output_dim=args.output_dim,
        implicit_hidden_dim=args.implicit_hidden_dim,
        hidden_dim=args.hidden_dim,
        is_low_rank=args.is_low_rank,
        rank=args.rank,
        kappa=args.kappa,
        mitr=args.mitr,
        grad_mitr=args.grad_mitr,
        tol=args.tol,
        grad_tol=args.grad_tol
        )

    print(f'Model size: {sum(p.numel() for p in model.parameters())} parameters')
    
    # Define loss function, optimizer, device and log directory
    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr)
    device = torch.device(f"cuda:{args.device}" if torch.cuda.is_available() else "cpu") 
    log_dir = os.path.join("results", f"im_rnn_{args.dataset}")
    
    # Train the model
    train_time_series(args, model, x_train, y_train, x_test, y_test, optimizer, loss_fn, log_dir, device)

if __name__ == '__main__':
    main()



from torchvision import datasets, transforms
from torch.utils.data import DataLoader
import torch

def load_data(args):

    # MNIST Dataset
    if args.dataset == 'mnist':
        transform = transforms.Compose([
            transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,)),
            transforms.Lambda(lambda x: torch.flatten(x))
            ])
        train_dataset = datasets.MNIST(root='./data', train=True, download=True, transform=transform)
        test_dataset = datasets.MNIST(root='./data', train=False, download=True, transform=transform)
        
    # CIFAR-10 Dataset
    elif args.dataset == 'cifar10':
        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            transforms.Lambda(lambda x: torch.flatten(x))  
        ])
        train_dataset = datasets.CIFAR10(root='./data', train=True, download=True, transform=transform)
        test_dataset = datasets.CIFAR10(root='./data', train=False, download=True, transform=transform)
    
    else:
        raise ValueError(f"Unsupported dataset: {args.dataset}")
    
    # Create DataLoaders
    train_loader = DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=args.batch_size, shuffle=False)
    
    return train_loader, test_loader
